#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
"""
This module contains a list of codecs for the DEV_ENCODED attribute type.
All codecs are based on the pair *format, data*. The format is a string
containing the codec signature and data is a sequence of bytes (string)
containing the encoded data.

This module contains a list of codecs capable of decoding several codings like
bz2, zip and json.

The :class:`CodecFactory` class allows you to get a codec object for a given
format and also to register new codecs.
The :class:`CodecPipeline` is a special codec that is able to code/decode a
sequence of codecs. This way you can have codecs 'inside' codecs.

Example::

    >>> from taurus.core.util.codecs import CodecFactory
    >>> cf = CodecFactory()
    >>> json_codec = cf.getCodec('json')
    >>> bz2_json_codec = cf.getCodec('bz2_json')
    >>> data = range(100000)
    >>> f1, enc_d1 = json_codec.encode(('', data))
    >>> f2, enc_d2 = bz2_json_codec.encode(('', data))
    >>> print len(enc_d1), len(enc_d2)
    688890 123511
    >>>
    >>> f1, dec_d1 = json_codec.decode((f1, enc_d1))
    >>> f2, dec_d2 = bz2_json_codec.decode((f2, enc_d2))

A Taurus related example::

    >>> # this example shows how to automatically get the data from a DEV_ENCODED attribute
    >>> import taurus
    >>> from taurus.core.util.codecs import CodecFactory
    >>> cf = CodecFactory()
    >>> devenc_attr = taurus.Attribute('a/b/c/devenc_attr')
    >>> v = devenc_attr.read()
    >>> codec = CodecFactory().getCodec(v.format)
    >>> f, d = codec.decode((v.format, v.value))
"""
from __future__ import absolute_import
from builtins import str, bytes

import copy

# need by VideoImageCodec
import struct
import sys
import numpy

from future.utils import (PY2, string_types)

from .singleton import Singleton
from .log import Logger
from .containers import CaselessDict

__all__ = ["Codec", "NullCodec", "ZIPCodec", "BZ2Codec", "JSONCodec",
           "Utf8Codec", "FunctionCodec", "PlotCodec", "CodecPipeline",
           "CodecFactory"]

__docformat__ = "restructuredtext"

if PY2:
    buffer_types = buffer, memoryview,
else:
    buffer_types = memoryview,


class Codec(Logger):
    """The base class for all codecs"""

    def __init__(self):
        """Constructor"""
        Logger.__init__(self, self.__class__.__name__)

    # TODO: similar code exists in encode and decode methods from different
    # codecs. It should be implemented in this base class 'Codec'.
    def encode(self, data, *args, **kwargs):
        """encodes the given data. This method is abstract an therefore must
        be implemented in the subclass.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :raises: NotImplementedError"""
        raise NotImplementedError("encode cannot be called on abstract Codec")

    def decode(self, data, *args, **kwargs):
        """decodes the given data. This method is abstract an therefore must
        be implemented in the subclass.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :raises: NotImplementedError"""
        raise NotImplementedError("decode cannot be called on abstract Codec")

    def __str__(self):
        return '%s()' % self.__class__.__name__

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class NullCodec(Codec):

    def encode(self, data, *args, **kwargs):
        """encodes with Null encoder. Just returns the given data

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        format = 'null'
        if len(data[0]):
            format += '_%s' % data[0]
        return format, data[1]

    def decode(self, data, *args, **kwargs):
        """decodes with Null encoder. Just returns the given data

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        if not data[0].startswith('null'):
            return data
        format = data[0].partition('_')[2]
        return format, data[1]


class ZIPCodec(Codec):
    """A codec able to encode/decode to/from gzip format. It uses the :mod:`zlib` module

    Example::

        >>> from taurus.core.util.codecs import CodecFactory

        >>> # first encode something
        >>> data = 100 * "Hello world\\n"
        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('zip')
        >>> format, encoded_data = codec.encode(("", data))
        >>> print len(data), len(encoded_data)
        1200, 31
        >>> format, decoded_data = codec.decode((format, encoded_data))
        >>> print decoded_data[20]
        'Hello world\\nHello wo'"""

    def encode(self, data, *args, **kwargs):
        """encodes the given data to gzip bytes. The given data **must** be bytes

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        import zlib
        format = 'zip'
        if len(data[0]):
            format += '_%s' % data[0]
        return format, zlib.compress(data[1])

    def decode(self, data, *args, **kwargs):
        """decodes the given data from a gzip bytes.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        import zlib
        if not data[0].startswith('zip'):
            return data
        format = data[0].partition('_')[2]
        return format, zlib.decompress(data[1])


class BZ2Codec(Codec):
    """A codec able to encode/decode to/from BZ2 format. It uses the :mod:`bz2` module

    Example::

        >>> from taurus.core.util.codecs import CodecFactory

        >>> # first encode something
        >>> data = 100 * "Hello world\\n"
        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('bz2')
        >>> format, encoded_data = codec.encode(("", data))
        >>> print len(data), len(encoded_data)
        1200, 68
        >>> format, decoded_data = codec.decode((format, encoded_data))
        >>> print decoded_data[20]
        'Hello world\\nHello wo'"""

    def encode(self, data, *args, **kwargs):
        """encodes the given data to bz2 bytes. The given data **must** be bytes

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        import bz2
        format = 'bz2'
        if len(data[0]):
            format += '_%s' % data[0]
        return format, bz2.compress(data[1])

    def decode(self, data, *args, **kwargs):
        """decodes the given data from bz2 bytes.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        import bz2
        if not data[0].startswith('bz2'):
            return data
        format = data[0].partition('_')[2]
        return format, bz2.decompress(data[1])


class PickleCodec(Codec):
    """A codec able to encode/decode to/from pickle format. It uses the
    :mod:`pickle` module.

    Example::

        >>> from taurus.core.util.codecs import CodecFactory

        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('pickle')
        >>>
        >>> # first encode something
        >>> data = { 'hello' : 'world', 'goodbye' : 1000 }
        >>> format, encoded_data = codec.encode(("", data))
        >>>
        >>> # now decode it
        >>> format, decoded_data = codec.decode((format, encoded_data))
        >>> print decoded_data
        {'hello': 'world', 'goodbye': 1000}"""

    def encode(self, data, *args, **kwargs):
        """encodes the given data to pickle bytes. The given data **must** be
        a python object that :mod:`pickle` is able to convert.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object"""
        import pickle
        format = 'pickle'
        if len(data[0]):
            format += '_%s' % data[0]
        # make it compact by default
        kwargs['protocol'] = kwargs.get('protocol', pickle.HIGHEST_PROTOCOL)
        return format, pickle.dumps(data[1], *args, **kwargs)

    def decode(self, data, *args, **kwargs):
        """decodes the given data from a pickle string.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object"""
        import pickle
        if not data[0].startswith('pickle'):
            return data
        format = data[0].partition('_')[2]

        if isinstance(data[1], buffer_types):
            data = data[0], bytes(data[1])

        return format, pickle.loads(data[1])


class JSONCodec(Codec):
    """A codec able to encode/decode to/from json format. It uses the
    :mod:`json` module.

    Example::

        >>> from taurus.core.util.codecs import CodecFactory

        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('json')
        >>>
        >>> # first encode something
        >>> data = { 'hello' : 'world', 'goodbye' : 1000 }
        >>> format, encoded_data = codec.encode(("", data))
        >>> print encoded_data
        '{"hello": "world", "goodbye": 1000}'
        >>>
        >>> # now decode it
        >>> format, decoded_data = codec.decode((format, encoded_data))
        >>> print decoded_data
        {'hello': 'world', 'goodbye': 1000}"""

    def encode(self, data, *args, **kwargs):
        """encodes the given data to a json string. The given data **must** be
        a python object that json is able to convert.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object"""
        import json
        format = 'json'
        if len(data[0]):
            format += '_%s' % data[0]
        # make it compact by default
        kwargs['separators'] = kwargs.get('separators', (',', ':'))
        return format, json.dumps(data[1], *args, **kwargs)

    def decode(self, data, *args, **kwargs):
        """decodes the given data from a json string.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object"""
        import json
        if not data[0].startswith('json'):
            return data
        format = data[0].partition('_')[2]

        ensure_ascii = kwargs.pop('ensure_ascii', False)

        if isinstance(data[1], buffer_types):
            data = data[0], str(data[1])
        
        data = json.loads(data[1])
        if ensure_ascii:
            data = self._transform_ascii(data)
        return format, data

    def _transform_ascii(self, data):
        if isinstance(data, string_types):
            return data.encode('utf-8')
        elif isinstance(data, dict):
            return self._transform_dict(data)
        elif isinstance(data, list):
            return self._transform_list(data)
        elif isinstance(data, tuple):
            return tuple(self._transform_list(data))
        else:
            return data

    def _transform_list(self, lst):
        return [self._transform_ascii(item) for item in lst]

    def _transform_dict(self, dct):
        newdict = {}
        for k, v in dct.items():
            newdict[self._transform_ascii(k)] = self._transform_ascii(v)
        return newdict


class Utf8Codec(Codec):
    """A codec able to encode/decode utf8 strings to/from bytes.
    Useful to adapt i/o encodings in a codec pipe.

    Example::

        >>> from taurus.core.util.codecs import CodecFactory

        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('zip_utf8_json')
        >>>
        >>> # first encode something
        >>> data = { 'hello' : 'world', 'goodbye' : 1000 }
        >>> format, encoded_data = codec.encode(("", data))
        >>>
        >>> # now decode it
        >>> _, decoded_data = codec.decode((format, encoded_data))
        >>> print decoded_data
    """

    def encode(self, data, *args, **kwargs):
        """
        Encodes the given utf8 string to bytes.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object
        """
        format = 'utf8'
        fmt, data = data
        if len(fmt):
            format += '_%s' % fmt
        return format, str(data).encode()

    def decode(self, data, *args, **kwargs):
        """decodes the given data from a bytes.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object
        """
        fmt, data = data
        fmt = fmt.partition('_')[2]
        return fmt, bytes(data).decode()


class BSONCodec(Codec):
    """A codec able to encode/decode to/from bson format. It uses the
    :mod:`bson` module.

    Example::

        >>> from taurus.core.util.codecs import CodecFactory

        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('bson')
        >>>
        >>> # first encode something
        >>> data = { 'hello' : 'world', 'goodbye' : 1000 }
        >>> format, encoded_data = codec.encode(("", data))
        >>>
        >>> # now decode it
        >>> _, decoded_data = codec.decode((format, encoded_data))
        >>> print decoded_data
        {'hello': 'world', 'goodbye': 1000}"""

    def encode(self, data, *args, **kwargs):
        """encodes the given data to a bson string. The given data **must** be
        a python object that bson is able to convert.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object"""
        import bson
        format = 'bson'
        if len(data[0]):
            format += '_%s' % data[0]
        return format, bson.BSON.encode(data[1], *args, **kwargs)

    def decode(self, data, *args, **kwargs):
        """decodes the given data from a bson string.

        :param data: (sequence[str, obj]) a sequence of two elements where the
                     first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the
                 first item is the encoding format of the second item object"""
        import bson
        if not data[0].startswith('bson'):
            return data
        format = data[0].partition('_')[2]
        ensure_ascii = kwargs.pop('ensure_ascii', False)

        data = data[0], bson.BSON(data[1])

        data = self.decode(data[1])
        if ensure_ascii:
            data = self._transform_ascii(data)
        return format, data

    def _transform_ascii(self, data):
        if isinstance(data, string_types):
            return data.encode('utf-8')
        elif isinstance(data, dict):
            return self._transform_dict(data)
        elif isinstance(data, list):
            return self._transform_list(data)
        elif isinstance(data, tuple):
            return tuple(self._transform_list(data))
        else:
            return data

    def _transform_list(self, lst):
        return [self._transform_ascii(item) for item in lst]

    def _transform_dict(self, dct):
        newdict = {}
        for k, v in dct.items():
            newdict[self._transform_ascii(k)] = self._transform_ascii(v)
        return newdict


class FunctionCodec(Codec):
    """A generic function codec"""

    def __init__(self, func_name):
        Codec.__init__(self)
        self._func_name = func_name

    def encode(self, data, *args, **kwargs):
        format = self._func_name
        if len(data[0]):
            format += '_%s' % data[0]
        return format, {'type': self._func_name, 'data': data[1]}

    def decode(self, data, *args, **kwargs):
        if not data[0].startswith(self._func_name):
            return data
        format = data[0].partition('_')[2]
        return format, data[1]


class PlotCodec(FunctionCodec):
    """A specialization of the :class:`FunctionCodec` for plot function"""

    def __init__(self):
        FunctionCodec.__init__(self, 'plot')


class VideoImageCodec(Codec):
    """A codec able to encode/decode to/from LImA video_image format.

    Example::

        >>> from taurus.core.util.codecs import CodecFactory
        >>> import PyTango

        >>> #first get an image from a LImA device to decode
        >>> data = PyTango.DeviceProxy(ccdName).read_attribute('video_last_image').value
        >>> cf = CodecFactory()
        >>> codec = cf.getCodec('VIDEO_IMAGE')
        >>> format,decoded_data = codec.decode(data)
        >>> # encode it again to check
        >>> format, encoded_data = codec.encode(("",decoded_data))
        >>> #compare images excluding the header:
        >>> data[1][32:] == encoded_data[32:]
        >>> #The headers can be different in the frameNumber
        >>> struct.unpack('!IHHqiiHHHH',data[1][:32])
        (1447314767, 1, 0, 6868, 1294, 964, 0, 32, 0, 0)
        >>> struct.unpack('!IHHqiiHHHH',encoded_data[:32])
        (1447314767, 1, 0, -1, 1294, 964, 0, 32, 0, 0)
    """

    VIDEO_HEADER_FORMAT = '!IHHqiiHHHH'

    def encode(self, data, *args, **kwargs):
        """encodes the given data to a LImA's video_image. The given data **must** be an numpy.array

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""

        # TODO: support encoding for colour imgage modes

        fmt = 'videoimage'
        if len(data[0]):
            fmt += '_%s' % data[0]
        # imgMode depends on numpy.array dtype
        imgMode = self.__getModeId(str(data[1].dtype))
        # frameNumber, unknown then -1
        height, width = data[1].shape
        header = self.__packHeader(imgMode, -1, width, height)
        buffer = data[1].tobytes()
        return fmt, header + buffer

    def decode(self, data, *args, **kwargs):
        """decodes the given data from a LImA's video_image.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""

        if data[0].startswith('VIDEO_IMAGE'):
            fixedformat = data[0].replace('VIDEO_IMAGE', 'videoimage')
            _, _, fmt = fixedformat.partition('_')
        elif data[0].startswith('videoimage'):
            _, _, fmt = data[0].partition('_')
        else:
            return data
        header = self.__unpackHeader(
            data[1][:struct.calcsize(self.VIDEO_HEADER_FORMAT)])

        imgBuffer = data[1][struct.calcsize(self.VIDEO_HEADER_FORMAT):]
        dtype = self.__getDtypeId(header['imageMode'])

        if header['imageMode'] == 6:
            # RGB24, 3 bytes per pixel
            rgba = numpy.frombuffer(imgBuffer, dtype)
            bbuf = rgba[0::3]
            gbuf = rgba[1::3]
            rbuf = rgba[2::3]
            r = rbuf.reshape(header['height'], header['width'])
            g = gbuf.reshape(header['height'], header['width'])
            b = bbuf.reshape(header['height'], header['width'])
            img2D = numpy.dstack((r, g, b))

        elif header['imageMode'] == 7:
            # RGBA 4 bytes per pixel
            rgba = numpy.frombuffer(imgBuffer, dtype)
            bbuf = rgba[0::4]
            gbuf = rgba[1::4]
            rbuf = rgba[2::4]
            #abuf = rgba[3::4]
            r = rbuf.reshape(header['height'], header['width'])
            g = gbuf.reshape(header['height'], header['width'])
            b = bbuf.reshape(header['height'], header['width'])
            #a = abuf.reshape(header['height'],header['width'])
            img2D = numpy.dstack((r, g, b))

        elif header['imageMode'] == 17:
            # YUV444 3 bytes per pixel
            yuv = numpy.frombuffer(imgBuffer, dtype)
            y = yuv[0::3]
            u = yuv[1::3]
            v = yuv[2::3]

            rbuff, gbuff, bbuff = self.__yuv2rgb(y, u, v)

            # Shape buffers to image size
            r = rbuff.reshape(header['height'], header['width'])
            g = gbuff.reshape(header['height'], header['width'])
            b = bbuff.reshape(header['height'], header['width'])

            # Build the RGB image
            img2D = numpy.dstack((r, g, b))

        elif header['imageMode'] == 16:
            # YUV422 4 bytes per 2 pixels
            yuv = numpy.frombuffer(imgBuffer, dtype)
            u = yuv[0::4]
            y1 = yuv[1::4]
            v = yuv[2::4]
            y2 = yuv[3::4]

            r1, g1, b1 = self.__yuv2rgb(y1, u, v)
            r2, g2, b2 = self.__yuv2rgb(y2, u, v)

            # Create RGB buffers
            rbuff = numpy.dstack((r1, r2)).reshape(
                header['height'] * header['width'])
            gbuff = numpy.dstack((g1, g2)).reshape(
                header['height'] * header['width'])
            bbuff = numpy.dstack((b1, b2)).reshape(
                header['height'] * header['width'])

            # Shape buffers to image size
            r = rbuff.reshape(header['height'], header['width'])
            g = gbuff.reshape(header['height'], header['width'])
            b = bbuff.reshape(header['height'], header['width'])

            # Build the RGB image
            img2D = numpy.dstack((r, g, b))

        elif header['imageMode'] == 15:
            # YUV411 6 bytes per 4 pixels
            yuv = numpy.frombuffer(imgBuffer, dtype)
            u = yuv[0::6]
            y1 = yuv[1::6]
            y2 = yuv[2::6]
            v = yuv[3::6]
            y3 = yuv[4::6]
            y4 = yuv[5::6]

            r1, g1, b1 = self.__yuv2rgb(y1, u, v)
            r2, g2, b2 = self.__yuv2rgb(y2, u, v)
            r3, g3, b3 = self.__yuv2rgb(y3, u, v)
            r4, g4, b4 = self.__yuv2rgb(y4, u, v)

            # Create RGB buffers
            rbuff = numpy.dstack((r1, r2, r3, r4)).reshape(
                header['height'] * header['width'])
            gbuff = numpy.dstack((g1, g2, g3, g4)).reshape(
                header['height'] * header['width'])
            bbuff = numpy.dstack((b1, b2, b3, b4)).reshape(
                header['height'] * header['width'])

            # Shape buffers to image size
            r = rbuff.reshape(header['height'], header['width'])
            g = gbuff.reshape(header['height'], header['width'])
            b = bbuff.reshape(header['height'], header['width'])

            img2D = numpy.dstack((r, g, b))

        else:
            img1D = numpy.frombuffer(imgBuffer, dtype)
            img2D = img1D.reshape(header['height'], header['width'])

        return fmt, img2D

    def __yuv2rgb(self, y, u, v):
        '''YUV444 to RGB888 conversion'''
        Cr = v - 128.0
        Cb = u - 128.0

        R = y + 1.402 * Cr
        G = y - 0.344 * Cb - 0.714 * Cr
        B = y + 1.772 * Cb

        return (numpy.clip(R, 0, 255), numpy.clip(G, 0, 255), numpy.clip(B, 0, 255))

    def __unpackHeader(self, header):
        h = struct.unpack(self.VIDEO_HEADER_FORMAT, header)
        headerDict = {}
        headerDict['magic'] = h[0]
        headerDict['headerVersion'] = h[1]
        headerDict['imageMode'] = h[2]
        headerDict['frameNumber'] = h[3]
        headerDict['width'] = h[4]
        headerDict['height'] = h[5]
        headerDict['endianness'] = h[6]
        headerDict['headerSize'] = h[7]
        headerDict['padding'] = h[8:]
        return headerDict

    def __packHeader(self, imgMode, frameNumber, width, height):
        magic = 0x5644454f
        version = 1
        endian = 0 if sys.byteorder == 'little' else 1
        hsize = struct.calcsize(self.VIDEO_HEADER_FORMAT)
        return struct.pack(self.VIDEO_HEADER_FORMAT,
                           magic,
                           version,
                           imgMode,
                           frameNumber,
                           width,
                           height,
                           endian,
                           hsize,
                           0, 0)  # padding

    def __getModeId(self, mode):
        return {  # when encode
            'uint8': 0,  # Core.Y8,
            'uint16': 1,  # Core.Y16,
            'uint32': 2,  # Core.Y32,
            'uint64': 3,  # Core.Y64,
            # when decode
            'Y8': 0,  # Core.Y8,
            'Y16': 1,  # Core.Y16,
            'Y32': 2,  # Core.Y32,
            'Y64': 3,  # Core.Y64,
            # TODO: other modes
            #'RGB555'     : 4,#Core.RGB555,
            #'RGB565'     : 5,#Core.RGB565,
            'RGB24': 6,  # Core.RGB24,
            'RGB32': 7,  # Core.RGB32,
            #'BGR24'      : 8,#Core.BGR24,
            #'BGR32'      : 9,#Core.BGR32,
            #'BAYER RG8'  : 10,#Core.BAYER_RG8,
            #'BAYER RG16' : 11,#Core.BAYER_RG16,
            #'BAYER BG8'  : 12,#Core.BAYER_BG8,
            #'BAYER BG16' : 13,#Core.BAYER_BG16,
            #'I420'       : 14,#Core.I420,
            #'YUV411'     : 15,#Core.YUV411,
            'YUV422': 16,  # Core.YUV422,
            #'YUV444'     : 17,#Core.YUV444
        }[mode]

    def __getFormatId(self, mode):
        return {0: 'B',
                1: 'H',
                2: 'I',
                3: 'L',
                #'RGB555'     : Core.RGB555,
                #'RGB565'     : Core.RGB565,
                6: 'RGB24',  # Core.RGB24,
                7: 'RGB32',  # Core.RGB32,
                # 8     : 'BGR24',#Core.BGR24,
                #'BGR32'      : Core.BGR32,
                #'BAYER RG8'  : Core.BAYER_RG8,
                #'BAYER RG16' : Core.BAYER_RG16,
                #'BAYER BG8'  : Core.BAYER_BG8,
                #'BAYER BG16' : Core.BAYER_BG16,
                #'I420'       : Core.I420,
                #'YUV411'     : Core.YUV411,
                16: 'YUV422',  # Core.YUV422,
                #'YUV444'     : Core.YUV444
                }[mode]

    def __getDtypeId(self, mode):
        return {0: 'uint8',
                1: 'uint16',
                2: 'uint32',
                3: 'uint64',
                #'RGB555'     : Core.RGB555,
                #'RGB565'     : Core.RGB565,
                6: 'uint8',  # Core.RGB24,
                7: 'uint8',  # Core.RGB32,
                #'BGR24'      : Core.BGR24,
                #'BGR32'      : Core.BGR32,
                #'BAYER RG8'  : Core.BAYER_RG8,
                #'BAYER RG16' : Core.BAYER_RG16,
                #'BAYER BG8'  : Core.BAYER_BG8,
                #'BAYER BG16' : Core.BAYER_BG16,
                #'I420'       : Core.I420,
                #'YUV411'     : Core.YUV411,
                16: 'uint8',  # Core.YUV422,
                #'YUV444'     : Core.YUV444
                }[mode]


class CodecPipeline(Codec, list):
    """The codec class used when encoding/decoding data with multiple encoders

    Example usage::

        >>> from taurus.core.util.codecs import CodecPipeline

        >>> data = range(100000)
        >>> codec = CodecPipeline('bz2_json')
        >>> format, encoded_data = codec.encode(("", data))

        # decode it
        format, decoded_data = codec.decode((format, encoded_data))
        print decoded_data"""

    def __init__(self, format):
        """Constructor. The CodecPipeline object will be created using
        the :class:`CodecFactory` to search for format(s) given in the format
        parameter.

        :param format: (str) a string representing the format."""

        Codec.__init__(self)
        list.__init__(self)

        f = CodecFactory()
        for i in format.split('_'):
            codec = f.getCodec(i)
            self.debug("Appending %s => %s" % (i, codec))
            if codec is None:
                raise TypeError(
                    'Unsupported codec %s (namely %s)' % (format, i))
            self.append(codec)
        self.debug("Done")

    def encode(self, data, *args, **kwargs):
        """encodes the given data.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        for codec in reversed(self):
            data = codec.encode(data, *args, **kwargs)
        return data

    def decode(self, data, *args, **kwargs):
        """decodes the given data.

        :param data: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object

        :return: (sequence[str, obj]) a sequence of two elements where the first item is the encoding format of the second item object"""
        for codec in self:
            data = codec.decode(data, *args, **kwargs)
        return data


class CodecFactory(Singleton, Logger):
    """The singleton CodecFactory class.

    To get the singleton object do::

        from taurus.core.util.codecs import CodecFactory
        f = CodecFactory()

    The :class:`CodecFactory` class allows you to get a codec object for a given
    format and also to register new codecs.
    The :class:`CodecPipeline` is a special codec that is able to code/decode a
    sequence of codecs. This way you can have codecs 'inside' codecs.

    Example::

        >>> from taurus.core.util.codecs import CodecFactory
        >>> cf = CodecFactory()
        >>> json_codec = cf.getCodec('json')
        >>> bz2_json_codec = cf.getCodec('bz2_json')
        >>> data = range(100000)
        >>> f1, enc_d1 = json_codec.encode(('', data))
        >>> f2, enc_d2 = bz2_json_codec.encode(('', data))
        >>> print len(enc_d1), len(enc_d2)
        688890 123511
        >>>
        >>> f1, dec_d1 = json_codec.decode((f1, enc_d1))
        >>> f2, dec_d2 = bz2_json_codec.decode((f2, enc_d2))

    A Taurus related example::

        >>> # this example shows how to automatically get the data from a DEV_ENCODED attribute
        >>> import taurus
        >>> from taurus.core.util.codecs import CodecFactory
        >>> cf = CodecFactory()
        >>> devenc_attr = taurus.Attribute('a/b/c/devenc_attr')
        >>> v = devenc_attr.read()
        >>> codec = CodecFactory().getCodec(v.format)
        >>> f, d = codec.decode((v.format, v.value))
    """

    #: Default minimum map of registered codecs
    CODEC_MAP = CaselessDict({
        'json': JSONCodec,
        'utf8': Utf8Codec,
        'bson': BSONCodec,
        'bz2': BZ2Codec,
        'zip': ZIPCodec,
        'pickle': PickleCodec,
        'plot': PlotCodec,
        'VIDEO_IMAGE': VideoImageCodec,  # deprecated
        'videoimage': VideoImageCodec,
        'null': NullCodec,
        'none': NullCodec,
        '': NullCodec})

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args, **kwargs):
        """Singleton instance initialization."""
        name = self.__class__.__name__
        self.call__init__(Logger, name)

        self._codec_klasses = copy.copy(CodecFactory.CODEC_MAP)

        # dict<str, Codec>
        # where:
        #  - key is the codec format
        #  - value is the codec object that supports the format
        self._codecs = CaselessDict()

    def registerCodec(self, format, klass):
        """Registers a new codec. If a codec already exists for the given format
        it is removed.

        :param format: (str) the codec id
        :param klass: (Codec class) the class that handles the format"""
        self._codec_klasses[format] = klass

        # del old codec if exists
        if format in self._codecs:
            del self._codecs[format]

    def unregisterCodec(self, format):
        """Unregisters the given format. If the format does not exist an exception
        is thrown.

        :param format: (str) the codec id

        :raises: KeyError"""
        if format in self._codec_klasses:
            del self._codec_klasses[format]

        if format in self._codecs:
            del self._codecs[format]

    def getCodec(self, format):
        """Returns the codec object for the given format or None if no suitable
        codec is found

        :param format: (str) the codec id

        :return: (Codec or None) the codec object for the given format"""
        codec = self._codecs.get(format)
        if codec is None:
            codec = self._getNewCodec(format)
            if not codec is None:
                self._codecs[format] = codec
        return codec

    def _getNewCodec(self, format):
        klass = self._codec_klasses.get(format)
        if not klass is None:
            ret = klass()
        else:
            try:
                ret = CodecPipeline(format)
            except:
                ret = self._codec_klasses.get('')()
        return ret

    def decode(self, data, *args, **kwargs):
        while len(data[0]):
            data = self.getCodec(data[0]).decode(data, *args, **kwargs)
        return data[1]

    def encode(self, format, data, *args, **kwargs):
        return self.getCodec(format).encode(data, *args, **kwargs)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""A pure Qt led widget"""

__all__ = ["LedColor", "LedStatus", "LedSize", "QLed", "QLedOld"]

__docformat__ = 'restructuredtext'

import sys

from taurus.external.qt import Qt
from taurus.core.util.enumeration import Enumeration
from taurus.qt.qtgui.icon import getCachedPixmap
from taurus.qt.qtgui.display.qpixmapwidget import QPixmapWidget

LedColor = Enumeration("LedColor",  [
                       "BLUE", "GREEN", "RED", "YELLOW", "ORANGE", "MAGENTA", "GRENOBLE", "BLACK", "WHITE"])
LedStatus = Enumeration("LedStatus", ["ON", "OFF"])
LedSize = Enumeration("LedSize",   [("SMALL", 24), ("LARGE", 48)])


class QLed(QPixmapWidget):
    """A Led"""
    DefaultLedPattern = "leds_images256:led_{color}_{status}.png"
    DefaultLedColor = "green"
    DefaultLedStatus = True
    DefaultLedInverted = False
    DefaultBlinkingInterval = 0

    def __init__(self, parent=None, designMode=False):
        self._ledStatus = self.DefaultLedStatus
        self._ledColor = self.DefaultLedColor
        self._ledPatternName = self.DefaultLedPattern
        self._ledInverted = self.DefaultLedInverted
        self._ledName = self.toLedName()
        self._timer = None
        QPixmapWidget.__init__(self, parent)
        self._refresh()

    def sizeHint(self):
        if self.layout() is None:
            return Qt.QSize(24, 24)
        return QPixmapWidget.sizeHint(self)

    def minimumSizeHint(self):
        """Overwrite the default minimum size hint (0,0) to be (16,16)
        :return: the minimum size hint 16,16
        :rtype: PyQt4.Qt.QSize"""
        return Qt.QSize(8, 8)

    def toLedName(self, status=None, color=None, inverted=None):
        """Gives the led name for the given status and color. If status or
        color are not given, the current led status or color are used.

        :param status: the status
        :type  status: bool
        :param color: the color
        :type  color: str
        :return: string containing the led name
        :rtype: str"""
        if status is None:
            status = self._ledStatus
        if color is None:
            color = self._ledColor
        if inverted is None:
            inverted = self._ledInverted
        if inverted:
            status = not status
        status = status and "on" or "off"
        return self._ledPatternName.format(color=color, status=status)

    def isLedColorValid(self, name):
        """Determines if the given color name is valid.
        :param color: the color
        :type  color: str
        :return: True is the given color name is valid or False otherwise
        :rtype: bool"""
        return name.upper() in LedColor

    def _refresh(self):
        """internal usage only"""
        self._ledName = self.toLedName()
        pixmap = getCachedPixmap(self._ledName)
        self.setPixmap(pixmap)
        return self.update()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'module': 'taurus.qt.qtgui.display',
            'group': 'Taurus Display',
            'icon': "designer:ledred.png",
            'container': False,
        }

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getLedPatternName(self):
        """Returns the current led pattern name
        :return: led pattern name
        :rtype: str"""
        return self._ledPatternName

    def setLedPatternName(self, name):
        """Sets the led pattern name. Should be a string containing a path
        to valid images. The string can contain the keywords:

            1. {status} - transformed to 'on' of 'off' according to the status
            2. {color} - transformed to the current led color

        Example: **:leds/images256/led_{color}_{status}.png** will be transformed
        to **:leds/images256/led_red_on.png** when the led status is True and
        the led color is red.

        :param name: new pattern
        :type  name: str"""
        self._ledPatternName = str(name)
        self._refresh()

    def resetLedPatternName(self):
        """Resets the led pattern to **:leds/images256/led_{color}_{status}.png**."""
        self.setLedPatternName(self.DefaultLedPattern)

    def getLedStatus(self):
        """Returns the led status
        :return: led status
        :rtype: bool"""
        return self._ledStatus

    def setLedStatus(self, status):
        """Sets the led status
        :param status: the new status
        :type  status: bool"""
        self._ledStatus = bool(status)
        self._refresh()

    def resetLedStatus(self):
        """Resets the led status"""
        self.setLedStatus(self.DefaultLedStatus)

    def toggleLedStatus(self):
        """toggles the current status of the led"""
        self.setLedStatus(not self.getLedStatus())

    def getLedInverted(self):
        """Returns if the led is inverted.
        :return: inverted mode
        :rtype: bool"""
        return self._ledInverted

    def setLedInverted(self, inverted):
        """Sets the led inverted mode
        :param status: the new inverted mode
        :type  status: bool"""
        self._ledInverted = bool(inverted)
        self._refresh()

    def resetLedInverted(self):
        """Resets the led inverted mode"""
        self.setLedInverted(self.DefaultLedInverted)

    def getLedColor(self):
        """Returns the led color
        :return: led color
        :rtype: str"""
        return self._ledColor

    def setLedColor(self, color):
        """Sets the led color
        :param status: the new color
        :type  status: str"""
        color = str(color).lower()
        if not self.isLedColorValid(color):
            raise Exception("Invalid color '%s'" % color)
        self._ledColor = color
        self._refresh()

    def resetLedColor(self):
        """Resets the led color"""
        self.setLedColor(self.DefaultLedColor)

    def setBlinkingInterval(self, interval):
        """sets the blinking interval (the time between status switching).
        Set to a nonpositive number for disabling blinking

        :param interval: (int) the blinking interval in millisecs. Set to 0 for disabling blinking
        """
        if interval > 0:
            if self._timer is None:
                self._timer = Qt.QTimer(self)
                self._timer.timeout.connect(self.toggleLedStatus)
            self._timer.start(interval)
        else:
            if self._timer is not None:
                self._timer.stop()
                self._timer = None

    def getBlinkingInterval(self):
        """returns the blinking interval

        :return: (int) blinking interval or 0 if blinking is not enabled.
        """
        if self._timer is None:
            return 0
        else:
            return self._timer.interval()

    def resetBlinkingInterval(self):
        """resets the blinking interval"""
        self.setBlinkingInterval(self.__class__.DefaultBlinkingInterval)

    #: This property holds the led status: False means OFF, True means ON
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QLed.getLedStatus`
    #:     * :meth:`QLed.setLedStatus`
    #:     * :meth:`QLed.resetLedStatus`
    ledStatus = Qt.pyqtProperty("bool", getLedStatus, setLedStatus,
                                resetLedStatus, doc="led status")

    #: This property holds the led color
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QLed.getLedColor`
    #:     * :meth:`QLed.setLedColor`
    #:     * :meth:`QLed.resetLedColor`
    ledColor = Qt.pyqtProperty("QString", getLedColor, setLedColor,
                               resetLedColor, doc="led color")

    #: This property holds the led inverted: False means do not invert the
    #. status, True means invert the status
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QLed.getLedInverted`
    #:     * :meth:`QLed.setLedInverted`
    #:     * :meth:`QLed.resetLedInverted`
    ledInverted = Qt.pyqtProperty("bool", getLedInverted, setLedInverted,
                                  resetLedInverted, doc="led inverted mode")

    #: This property holds the led pattern name
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QLed.getLedPatternName`
    #:     * :meth:`QLed.setLedPatternName`
    #:     * :meth:`QLed.resetLedPatternName`
    ledPattern = Qt.pyqtProperty("QString", getLedPatternName, setLedPatternName,
                                 resetLedPatternName, doc="led pattern name")

    #: This property holds the blinking interval in millisecs. 0 means no blinking
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QLed.getBlinkingInterval`
    #:     * :meth:`QLed.setBlinkingInterval`
    #:     * :meth:`QLed.resetBlinkingInterval`
    blinkingInterval = Qt.pyqtProperty("int", getBlinkingInterval, setBlinkingInterval,
                                       resetBlinkingInterval, doc="blinking interval")


class QLedOld(Qt.QLabel):
    ledDirPattern = ":leds/images%(size)d"

    def __init__(self, parent=None, ledsize=LedSize.SMALL,
                 ledcolor=LedColor.GREEN):
        Qt.QLabel.__init__(self, parent)

        self.ledsize = ledsize
        self.status = LedStatus.OFF
        self.ledcolor = ledcolor

        self.setObjectName("Led")
        sizePolicy = Qt.QSizePolicy(
            Qt.QSizePolicy.Policy(0), Qt.QSizePolicy.Policy(0))
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(sizePolicy)

        # no need to call changeColor. changeSize calls it for us
        self.changeSize(ledsize)

        self.retranslateUi(self)
        Qt.QMetaObject.connectSlotsByName(self)

    def toLedName(self, status, color):
        name = "led" + LedColor.whatis(color).lower()
        if status == LedStatus.OFF:
            name += LedStatus.whatis(status).lower()
        return name

    def toCompleteLedName(self, size, status, color):
        ret = self.ledDirPattern % {"size": size}
        ret += "/" + self.toLedName(status, color) + ".png"
        return ret

    def tr(self, string):
        return Qt.QApplication.translate("Led", string, None, Qt.QApplication.UnicodeUTF8)

    def retranslateUi(self, Led):
        Led.setWindowTitle(self.tr("Form"))

    def on(self):
        if self.status == LedStatus.OFF:
            self.status = LedStatus.ON
            self.setPixmap(self.onled)

    def off(self):
        if self.status == LedStatus.ON:
            self.status = LedStatus.OFF
            self.setPixmap(self.offled)

    def changeSize(self, ledSize):
        self.ledsize = ledSize
        self.resize(ledSize, ledSize)
        self.setMinimumSize(Qt.QSize(ledSize, ledSize))
        self.setMaximumSize(Qt.QSize(ledSize, ledSize))

        # force a call to led color to change the place from which the images
        # are fetched
        self.changeColor(self.ledcolor)

    def changeColor(self, ledColor):
        self.ledcolor = ledColor

        off_pixmap_name = self.toCompleteLedName(
            self.ledsize, LedStatus.OFF, ledColor)
        self.offled = Qt.QPixmap(off_pixmap_name)
        on_pixmap_name = self.toCompleteLedName(
            self.ledsize, LedStatus.ON, ledColor)
        self.onled = Qt.QPixmap(on_pixmap_name)

        if self.status == LedStatus.OFF:
            self.setPixmap(self.offled)
        else:
            self.setPixmap(self.onled)

    def _setProblemsBackground(self, doIt=True):
        if doIt:
            self.setAutoFillBackground(True)
            bg_brush = Qt.QBrush()
            bg_brush.setStyle(Qt.Qt.BDiagPattern)
            palette = self.palette()
            palette.setBrush(Qt.QPalette.Window, bg_brush)
            palette.setBrush(Qt.QPalette.Base, bg_brush)
        else:
            self.setAutoFillBackground(False)


def main():
    """hello"""

    app = Qt.QApplication.instance()
    owns_app = app is None

    if owns_app:
        from taurus.qt.qtgui.application import TaurusApplication
        app = TaurusApplication(sys.argv, cmd_line_parser=None)

    w = Qt.QWidget()
    layout = Qt.QGridLayout()
    layout.setContentsMargins(2, 2, 2, 2)
    layout.setSpacing(2)
    w.setLayout(layout)
    for i, color in enumerate(LedColor.keys()):
        led = QLed()
        led.ledColor = color
        led.ledStatus = True
        layout.addWidget(led, i, 0)
        led = QLed()
        led.ledColor = color
        led.ledStatus = False
        layout.addWidget(led, i, 1)
        led = QLed()
        led.ledColor = color
        led.blinkingInterval = 500
        layout.addWidget(led, i, 2)
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurusdevicepropertytable.py:
"""

# todo: tango-centric

from __future__ import print_function
from future.utils import string_types
from builtins import str

from taurus.external.qt import Qt, QtCore, QtGui
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.core.taurusdevice import TaurusDevice


__all__ = ["TaurusPropTable"]


class TaurusPropTable(QtGui.QTableWidget, TaurusBaseWidget):
    '''
    This widget will show a list of properties of device and the list of values.
    @todo add a frame for Add, Delete and Refresh buttons!
    '''
    # TODO This widget is Tango-centric

    def __init__(self, parent=None, designMode=False):
        try:
            name = "TaurusPropTable"
            self._useParentModel = True
            self._localModel = ''
            self.call__init__wo_kw(QtGui.QTableWidget, parent)
            self.call__init__(TaurusBaseWidget, name, designMode=designMode)
            self.setObjectName(name)
            self.defineStyle()
            self.db = None

        except Exception as e:
            self.traceback()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    def setModel(self, model):
        TaurusBaseWidget.setModel(self, model)
        # Update the table
        modelObj = self.getModelObj()
        if modelObj is None:
            dev_name = None
            self.db = None
        else:
            validator = modelObj.getNameValidator()
            groups = validator.getUriGroups(modelObj.getFullName())
            dev_name = groups['devname']
            # -----------------------------------------------------------------
            # Workaround for bug-256
            # (use next line instead of the other one when bug-256 is fixed)
            # self.db = modelObj.getParentObj() # use this instead
            self.db = modelObj.factory().getAuthority(groups['authority'])
            # -----------------------------------------------------------------
        self.setTable(dev_name)

    def sizeHint(self):
        return QtGui.QTableWidget.sizeHint(self)

    def minimumSizeHint(self):
        return QtGui.QTableWidget.minimumSizeHint(self)

    def getModelClass(self):
        return TaurusDevice

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.table'
        ret['group'] = 'Taurus Views'
        ret['icon'] = "designer:table.png"
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = QtCore.pyqtProperty("QString", TaurusBaseWidget.getModel,
                                setModel,
                                TaurusBaseWidget.resetModel)

    useParentModel = QtCore.pyqtProperty("bool",
                                         TaurusBaseWidget.getUseParentModel,
                                         TaurusBaseWidget.setUseParentModel,
                                         TaurusBaseWidget.resetUseParentModel)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # My methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @QtCore.pyqtSlot('QString')
    def setTable(self, dev_name):
        """
        Fills the table with the names of properties and their values for the
        given device.
        """
        if dev_name is None:
            return  # TODO: the table should be cleaned

        elif self.db is None:
            self.warning('Model must be set before calling setTable')
            return
        try: 
            self.cellChanged.disconnect(self.valueChanged)
        except: 
            pass

        dev_name = str(dev_name)
        self.list_prop = list(self.db.get_device_property_list(dev_name, '*'))
        self.setRowCount(len(self.list_prop))
        for i in range(0, len(self.list_prop)):
            elem = self.list_prop[i]
            self.setText(elem, i, 0)
            self.dictionary = self.db.get_device_property(dev_name,
                                                          self.list_prop)
            self.debug('Getting %s properties: %s -> %s',
                       dev_name, self.list_prop, self.dictionary)
            value = self.dictionary[elem]
            self.debug('TaurusPropsTable: property %s is type %s',
                       elem, type(value))
            USE_TABLES = False
            if USE_TABLES:
                self.setPropertyValue(value, i, 1)
            else:
                if not isinstance(value, string_types):  # not something like an string
                    # adding new lines in between elements in the list
                    value = '\n'.join(str(v) for v in value)
                self.setText(str(value), i, 1)

        self.updateStyle()
        self.dev_name = dev_name
        self.setWindowTitle('%s Properties' % dev_name)
        self.resizeColumnsToContents()
        self.resizeRowsToContents()
        # Signals @todo
        # Commented as it didn't work really well (many pop-ups open!)
        # QtCore.QObject.connect(self,QtCore.SIGNAL("cellDoubleClicked(int,int)"),self.valueDoubleClicked)
        # QtCore.QObject.connect(self,QtCore.SIGNAL("cellChanged(int,int)"),self.valueChanged)

    def defineStyle(self):
        """ Defines the initial style for the widget """
        self.setWindowTitle('Properties')
        self.setColumnCount(2)
        self.setRowCount(0)
        self.setGeometry(QtCore.QRect(0, 0, 400, 500))

        self.setColumnWidth(0, 124)
        self.setColumnWidth(1, 254)

        headerItem = QtGui.QTableWidgetItem()
        headerItem.setText(QtGui.QApplication.translate(
            "PLCTabWidget", "Property Name", None, QtGui.QApplication.UnicodeUTF8))
        self.setHorizontalHeaderItem(0, headerItem)

        headerItem1 = QtGui.QTableWidgetItem()
        headerItem1.setText(QtGui.QApplication.translate(
            "PLCTabWidget", "Value", None, QtGui.QApplication.UnicodeUTF8))
        self.setHorizontalHeaderItem(1, headerItem1)
        hh = self.horizontalHeader()
        if hh.length() > 0:
            try:
                hh.setSectionResizeMode(hh.ResizeToContents)
            except AttributeError:  # PyQt4
                hh.setResizeMode(hh.ResizeToContents)

    def updateStyle(self):
        self.resizeRowsToContents()
        hh = self.horizontalHeader()
        if hh.length() > 0:
            try:
                hh.setSectionResizeMode(hh.ResizeToContents)
            except AttributeError:  # PyQt4
                hh.setResizeMode(hh.ResizeToContents)
            # self.resizeColumnsToContents()

    def contextMenuEvent(self, event):
        ''' This function is called when right clicking on qwt plot area. A pop up menu will be
        shown with the available options. '''
        self.info('TaurusPropTable.contextMenuEvent()')
        menu = Qt.QMenu(self)
        configDialogAction = menu.addAction("Add new property")
        configDialogAction.triggered.connect(self.addProperty)
        configDialogAction = menu.addAction("Delete property")
        configDialogAction.triggered.connect(self.deleteProperty)
        configDialogAction = menu.addAction("Edit property")
        configDialogAction.triggered.connect(self.editProperty)
        menu.addSeparator()
        menu.exec_(event.globalPos())
        del menu

    def setText(self, value, i=None, j=None):
        item = QtGui.QTableWidgetItem()
        item.setFlags(Qt.Qt.ItemIsEnabled)
        item.setText(QtGui.QApplication.translate("PLCTabWidget",
                                                  value, None, QtGui.QApplication.UnicodeUTF8))
        if i is not None and j is not None:
            self.setItem(i, j, item)
        # TODO: the info does not change with the events.

    def valueDoubleClicked(self, x, y):
        self.info('TaurusPropTable.valueDoubleClicked(%s,%s)' % (x, y))
        # opens a dialog for multiline edition
        self.editProperty()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # @name Property Edition
    # @{
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def addProperty(self):
        text, ok = QtGui.QInputDialog.getText(
            self, 'New Property', 'Property name:')
        if ok:
            text1 = str(text)
            new_prop_name = str(text1)
            new_prop_value = '0'
            dict1 = {new_prop_name: [new_prop_value]}
            self.db.put_device_property(self.dev_name, dict1)
            self.setTable(self.dev_name)

    def deleteProperty(self):
        row = self.currentRow()
        prop_name = self.item(row, 0).text()
        list = [str(prop_name)]
        yes = QtGui.QMessageBox.Ok
        no = QtGui.QMessageBox.Cancel
        result = QtGui.QMessageBox.question(
            self, "Removing property", "Would you like to delete property  '" + prop_name + "'  ?", yes, no)
        if result == yes:
            self.db.delete_device_property(self.dev_name, list)
            self.setTable(self.dev_name)

    def editProperty(self):
        row = self.currentRow()
        col = self.currentColumn()

        item1 = QtGui.QTableWidgetItem()
        item1 = self.item(row, 0)
        prop_name = item1.text()
        prop_name = str(prop_name)
        self.prop_name2 = prop_name
        self.info('TaurusPropsTable.editProperty(%s)' % prop_name)

        item2 = QtGui.QTableWidgetItem()
        item2 = self.item(row, 1)
        prop_value = item2.text()
        prop_value = str(prop_value)

        if col == 0:
            new_text, ok = QtGui.QInputDialog.getText(
                self, 'Rename', 'Write new name of property:')
            if ok:
                new_text = str(new_text)
                new_text = str(new_text)
                list = [prop_name]
                dict = {new_text: [prop_value]}
                # usuwanie musze umiescic gdzies wczesniej bo inaczej usuwam
                # juz zmieniana nazwe z listy property
                self.db.delete_device_property(self.dev_name, list)
                self.db.put_device_property(self.dev_name, dict)
                self.setTable(self.dev_name)
        elif col == 1:
            # Create the dilog to edit multiply text
            dialogx = EditTextDialog(self)
            dialogx.setText(prop_value)
            dialogx.exec_()
            ok = dialogx.getResult()  # OK or Cancel
            if ok:
                new_text = dialogx.getNewText()  # method of dialog to get the changed text
                self.setNewPropertyValue(new_text)

    def setNewPropertyValue(self, new_text):
        new_text = str(new_text)
        new_text = str(new_text)
        values = {self.prop_name2: new_text.replace('\r', '').split('\n')}
        self.db.put_device_property(self.dev_name, values)
        self.setTable(self.dev_name)
        self.updateStyle()
    # @}

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Methods  for database commands
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def get_device_property_names(self, dev_name, wildcard='*'):
        return self.db.get_device_property_list(dev_name, wildcard)

    def put_device_property(self, dev_name, dict):
        return self.db.put_device_property(dev_name, dict)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # @name DEPRECATED METHODS
    # @{
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def valueChanged(self):
        ''' @deprecated valueChanged THIS DOES NOTHING! '''
        row = self.currentRow()
        #item = QtGui.QTableWidgetItem()
        item = self.item(row, 0)
        prop_name = item.text()
        prop_name = str(prop_name)
        #item_2 = QtGui.QTableWidgetItem()
        item_2 = self.item(row, 1)
        prop_value = item_2.text()
        prop_value = str(prop_value)
        dict = {prop_name: [prop_value]}
        list = [self.prop_name]
        self.db.delete_device_property(self.dev_name, list)
        self.db.put_device_property(self.dev_name, dict)

    def setPropertyValue(self, value, i, j):
        ''' This method inserts a new table widget inside the cell
        @deprecated ... use setText() and editProperty() event call instead!!!
        '''
        if len(value) == 1 and isinstance(value[0], string_types):
            value = value[0]
        if isinstance(value, string_types):  # and '\n' in value:
            value = value.split('\n')
        if False:  # isinstance(value,str):
            item = QtGui.QTableWidgetItem()
            item.setText(QtGui.QApplication.translate(
                "PLCTabWidget", value, None, QtGui.QApplication.UnicodeUTF8))
            self.setItem(i, j, item)
        else:
            item = QtGui.QTableWidget(len(value), 1)
            item.setItemDelegate(Delegate(item))
            item.horizontalHeader().hide()
            item.verticalHeader().hide()
            # item.setGridStyle(Qt.Qt.DashLine)
            for k, v in enumerate(value):
                cell = QtGui.QTableWidgetItem()
                cell.setText(QtGui.QApplication.translate(
                    "PLCTabWidget", v, None, QtGui.QApplication.UnicodeUTF8))
                item.setItem(k, 0, cell)
            item.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding,
                               QtGui.QSizePolicy.MinimumExpanding)
            self.setCellWidget(i, j, item)
            self.resizeColumnsToContents()
            self.resizeRowsToContents()
        return

    # @}


class EditTextDialog(QtGui.QDialog):
    """ This class create the dialog using to edit multiply text """

    def __init__(self, parent=None):
        print ('In EditTextDialog.__init__()')
        QtGui.QDialog.__init__(self, parent)
        self.setModal(1)
        self.initComponents()
        self.show()
        self.result = 0
        # Signals

        self.buttonBox.accepted.connect(self.pressOK)
        self.buttonBox.rejected.connect(self.close)

    def initComponents(self):
        widgetLayout = QtGui.QVBoxLayout(self)
        widgetLayout.setContentsMargins(10, 10, 10, 10)
        self.editText = QtGui.QTextEdit()
        self.buttonBox = QtGui.QDialogButtonBox()
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(
            QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.NoButton | QtGui.QDialogButtonBox.Ok)
        widgetLayout.addWidget(self.editText)
        widgetLayout.addWidget(self.buttonBox)

    def pressOK(self):
        print ('In EditTextDialog.pressOk()')
        self.new_text = self.editText.toPlainText()
        self.result = 1
        self.close()

    def setText(self, text):
        self.editText.setText(text)
        self.new_text = text

    def getNewText(self):
        return self.new_text

    def getResult(self):
        return self.result


class Delegate(QtGui.QItemDelegate):
    ''' Usage:
            table = QtGui.QTableWidget()
        table.setItemDelegate(Delegate(table))
    '''

    def __init__(self, parent=None):
        QtGui.QItemDelegate.__init__(self, parent)

    def sizeHint(self, option, index):
        table = self.parent()
        widget = table.cellWidget(index.row(), index.column())
        if not widget:
            widget = table.itemAt(index.row(), index.column())
        size = widget.sizeHint()
        return size

if __name__ == '__main__':
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(app_name="TaurusDevice property table",
                            cmd_line_parser=None)
    widget = TaurusPropTable()
    args = sys.argv
    if len(args) == 1:
        model = 'sys/tg_test/1'
    else:
        model = str(args[1])
    widget.setModel(model)
    widget.show()
    app.exec_()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Generic tests (instantiation, set models, etc) for several widgets
"""

import pytest
from importlib import import_module
import taurus
from taurus.test.pytest import check_taurus_deprecations


DB = 'sys/database/2'
TGT = 'tango:sys/tg_test/1'
TGT_FL_SC = TGT + '/float_scalar'
TGT_FL_SP = TGT + '/float_spectrum'
TGT_SH_SC = TGT + '/short_scalar'
TGT_UC_IM = TGT + '/uchar_image_ro'
TGT_ST_SP = TGT + '/string_spectrum'
TGT_WAVE = TGT + '/wave'
TGT_NOVAL = TGT + '/no_value'
TGT_EXC = TGT + '/throw_exception'
EV_INT = 'eval:123'
EV_Q = 'eval:1.23*UR.mV'
EV_RND5 = 'eval:rand(5)'

# TODO: create and keep model objects in setup_module to speed up


def _import_obj(obj_str, package="taurus.qt.qtgui"):
    """
    returns objects described by a string like "<modname>[:<objname>]"
    """
    if ":" in obj_str:
        modname, oname = obj_str.split(":")
        return getattr(import_module(modname, package=package), oname)
    else:
        return import_module(obj_str, package=package)


@pytest.mark.parametrize(
    "widgetname,depr,models",
    [
        (".display:TaurusLabel", 0, [TGT_WAVE, "", EV_INT, None]),
        (".button:TaurusCommandButton", 0, [TGT, "", DB, None]),
        (".panel:TaurusAttrForm", 0, [TGT, "", DB, None]),
        (".panel:TaurusForm", 0, [
                  [TGT],
                  [TGT_WAVE],
                  [],
                  "",
                  [EV_INT],
                  None,
                  [TGT_SH_SC, TGT_UC_IM, TGT_ST_SP, TGT_NOVAL, TGT_EXC],
                  [""],
                  ",".join((TGT, EV_INT)),
                  " ".join((TGT_UC_IM, EV_RND5)),
                  [None]
        ]),
    ]
)
def test_set_models(qtbot, caplog, widgetname, depr, models):
    """
    Generic test that checks that a widget can be instantiated and given
    its setModel called sequentially.

    It can be parameterized or run with functools.partial
    """
    with check_taurus_deprecations(caplog, expected=depr):
        klass = _import_obj(widgetname)
        w = klass()
        qtbot.addWidget(w)

        for model in models:
            if not model:
                model_obj = None
            else:
                try:
                    model_obj = taurus.Object(model)
                except:  # allow non-string models (e.g. lists of models)
                    model_obj = None
            w.setModel(model)
            assert w.getModelObj() == model_obj



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
scales.py: Custom scales used by taurus.qt.qtgui.plot module
"""

from __future__ import print_function

import numpy
from datetime import datetime, timedelta
from time import mktime
from taurus.external.qt import Qt, Qwt5


__all__ = ["DateTimeScaleEngine", "DeltaTimeScaleEngine", "FixedLabelsScaleEngine",
           "FancyScaleDraw", "TaurusTimeScaleDraw", "DeltaTimeScaleDraw",
           "FixedLabelsScaleDraw"]


def _getDefaultAxisLabelsAlignment(axis, rotation):
    '''return a "smart" alignment for the axis labels depending on the axis
    and the label rotation

    :param axis: (Qwt5.QwtPlot.Axis) the axis
    :param rotation: (float) The rotation (in degrees, clockwise-positive)

    :return: (Qt.Alignment) an alignment
    '''
    if axis == Qwt5.QwtPlot.xBottom:
        if rotation == 0:
            return Qt.Qt.AlignHCenter | Qt.Qt.AlignBottom
        elif rotation < 0:
            return Qt.Qt.AlignLeft | Qt.Qt.AlignBottom
        else:
            return Qt.Qt.AlignRight | Qt.Qt.AlignBottom
    elif axis == Qwt5.QwtPlot.yLeft:
        if rotation == 0:
            return Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter
        elif rotation < 0:
            return Qt.Qt.AlignLeft | Qt.Qt.AlignBottom
        else:
            return Qt.Qt.AlignLeft | Qt.Qt.AlignTop
    elif axis == Qwt5.QwtPlot.yRight:
        if rotation == 0:
            return Qt.Qt.AlignRight | Qt.Qt.AlignVCenter
        elif rotation < 0:
            return Qt.Qt.AlignRight | Qt.Qt.AlignTop
        else:
            return Qt.Qt.AlignRight | Qt.Qt.AlignBottom
    elif axis == Qwt5.QwtPlot.xTop:
        if rotation == 0:
            return Qt.Qt.AlignHCenter | Qt.Qt.AlignTop
        elif rotation < 0:
            return Qt.Qt.AlignLeft | Qt.Qt.AlignTop
        else:
            return Qt.Qt.AlignRight | Qt.Qt.AlignTop


class FancyScaleDraw(Qwt5.QwtScaleDraw):

    '''This is a scaleDraw with a tuneable palette and label formats'''

    def __init__(self, format=None, palette=None):
        Qwt5.QwtScaleDraw.__init__(self)
        self._labelFormat = format
        self._palette = palette

    def setPalette(self, palette):
        '''pass a QPalette or None to use default'''
        self._palette = palette

    def getPalette(self):
        return self._palette

    def setLabelFormat(self, format):
        '''pass a format string (e.g. "%g") or None to use default (it uses the locale)'''
        self._labelFormat = format
        self.invalidateCache()  # to force repainting of the labels

    def getLabelFormat(self):
        '''pass a format string (e.g. "%g") or None to use default (it uses the locale)'''
        return self._labelFormat

    def label(self, val):
        if str(self._labelFormat) == "":
            return Qwt5.QwtText()
        if self._labelFormat is None:
            return Qwt5.QwtScaleDraw.label(self, val)
        else:
            return Qwt5.QwtText(self._labelFormat % val)

    def draw(self, painter, palette):
        if self._palette is None:
            Qwt5.QwtScaleDraw.draw(self, painter, palette)
        else:
            Qwt5.QwtScaleDraw.draw(self, painter, self._palette)


class DateTimeScaleEngine(Qwt5.QwtLinearScaleEngine):

    def __init__(self, scaleDraw=None):
        Qwt5.QwtLinearScaleEngine.__init__(self)
        self.setScaleDraw(scaleDraw)

    def setScaleDraw(self, scaleDraw):
        self._scaleDraw = scaleDraw

    def scaleDraw(self):
        return self._scaleDraw

    def divideScale(self, x1, x2, maxMajSteps, maxMinSteps, stepSize):
        ''' Reimplements Qwt5.QwtLinearScaleEngine.divideScale

        **Important**: The stepSize parameter is **ignored**.

        :return: (Qwt5.QwtScaleDiv) a scale division whose ticks are aligned with
                 the natural time units '''

        # if stepSize != 0:
        #    scaleDiv = Qwt5.QwtLinearScaleEngine.divideScale(self, x1, x2, maxMajSteps, maxMinSteps, stepSize)
        #    scaleDiv.datetimeLabelFormat = "%Y/%m/%d %H:%M%S.%f"
        #    return scaleDiv

        interval = Qwt5.QwtDoubleInterval(x1, x2).normalized()
        if interval.width() <= 0:
            return Qwt5.QwtScaleDiv()

        dt1 = datetime.fromtimestamp(interval.minValue())
        dt2 = datetime.fromtimestamp(interval.maxValue())

        if dt1.year < 1900 or dt2.year > 9999:  # limits in time.mktime and datetime
            return Qwt5.QwtScaleDiv()

        majticks = []
        medticks = []
        minticks = []

        dx = interval.width()

        # = 3600s*24*(365+366) = 2 years (counting a leap year)
        if dx > 63072001:
            format = "%Y"
            for y in range(dt1.year + 1, dt2.year):
                dt = datetime(year=y, month=1, day=1)
                majticks.append(mktime(dt.timetuple()))

        elif dx > 5270400:  # = 3600s*24*61 = 61 days
            format = "%Y %b"
            d = timedelta(days=31)
            dt = dt1.replace(day=1, hour=0, minute=0,
                             second=0, microsecond=0) + d
            while(dt < dt2):
                # make sure that we are on day 1 (even if always sum 31 days)
                dt = dt.replace(day=1)
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 172800:  # 3600s24*2 = 2 days
            format = "%b/%d"
            d = timedelta(days=1)
            dt = dt1.replace(hour=0, minute=0, second=0, microsecond=0) + d
            while(dt < dt2):
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 7200:  # 3600s*2 = 2hours
            format = "%b/%d-%Hh"
            d = timedelta(hours=1)
            dt = dt1.replace(minute=0, second=0, microsecond=0) + d
            while(dt < dt2):
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 1200:  # 60s*20 =20 minutes
            format = "%H:%M"
            d = timedelta(minutes=10)
            dt = dt1.replace(minute=(dt1.minute // 10) * 10,
                             second=0, microsecond=0) + d
            while(dt < dt2):
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 120:  # =60s*2 = 2 minutes
            format = "%H:%M"
            d = timedelta(minutes=1)
            dt = dt1.replace(second=0, microsecond=0) + d
            while(dt < dt2):
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 20:  # 20 s
            format = "%H:%M:%S"
            d = timedelta(seconds=10)
            dt = dt1.replace(second=(dt1.second // 10) * 10, microsecond=0) + d
            while(dt < dt2):
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 2:  # 2s
            format = "%H:%M:%S"
            majticks = list(range(int(x1) + 1, int(x2)))

        else:  # less than 2s (show microseconds)
            scaleDiv = Qwt5.QwtLinearScaleEngine.divideScale(
                self, x1, x2, maxMajSteps, maxMinSteps, stepSize)
            self.scaleDraw().setDatetimeLabelFormat("%S.%f")
            return scaleDiv

        # make sure to comply with maxMajTicks
        L = len(majticks)
        if L > maxMajSteps:
            majticks = majticks[::int(numpy.ceil(float(L) / maxMajSteps))]

        scaleDiv = Qwt5.QwtScaleDiv(interval, minticks, medticks, majticks)
        self.scaleDraw().setDatetimeLabelFormat(format)
        if x1 > x2:
            scaleDiv.invert()

        # START DEBUG
        # print "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
        # for tk in  scaleDiv.ticks(scaleDiv.MajorTick):
        #    print datetime.fromtimestamp(tk).isoformat()
        # print "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
        # END DEBUG

        return scaleDiv

    @staticmethod
    def getDefaultAxisLabelsAlignment(axis, rotation):
        '''return a "smart" alignment for the axis labels depending on the axis
        and the label rotation

        :param axis: (Qwt5.QwtPlot.Axis) the axis
        :param rotation: (float) The rotation (in degrees, clockwise-positive)

        :return: (Qt.Alignment) an alignment
        '''
        return _getDefaultAxisLabelsAlignment(axis, rotation)

    @staticmethod
    def enableInAxis(plot, axis, scaleDraw=None, rotation=None):
        '''convenience method that will enable this engine in the given
        axis. Note that it changes the ScaleDraw as well.

        :param plot: (Qwt5.QwtPlot) the plot to change
        :param axis: (Qwt5.QwtPlot.Axis) the id of the axis
        :param scaleDraw: (Qwt5.QwtScaleDraw) Scale draw to use. If None given,
                          the current ScaleDraw for the plot will be used if
                          possible, and a :class:`TaurusTimeScaleDraw` will be set if not
        :param rotation: (float or None) The rotation of the labels (in degrees, clockwise-positive)
        '''
        if scaleDraw is None:
            scaleDraw = plot.axisScaleDraw(axis)
            if not isinstance(scaleDraw, TaurusTimeScaleDraw):
                scaleDraw = TaurusTimeScaleDraw()
        plot.setAxisScaleDraw(axis, scaleDraw)
        plot.setAxisScaleEngine(axis, DateTimeScaleEngine(scaleDraw))
        if rotation is not None:
            alignment = DateTimeScaleEngine.getDefaultAxisLabelsAlignment(
                axis, rotation)
            plot.setAxisLabelRotation(axis, rotation)
            plot.setAxisLabelAlignment(axis, alignment)

    @staticmethod
    def disableInAxis(plot, axis, scaleDraw=None, scaleEngine=None):
        '''convenience method that will disable this engine in the given
        axis. Note that it changes the ScaleDraw as well.

        :param plot: (Qwt5.QwtPlot) the plot to change
        :param axis: (Qwt5.QwtPlot.Axis) the id of the axis
        :param scaleDraw: (Qwt5.QwtScaleDraw) Scale draw to use. If None given,
                          a :class:`FancyScaleDraw` will be set
        :param scaleEngine: (Qwt5.QwtScaleEngine) Scale draw to use. If None given,
                          a :class:`Qwt5.QwtLinearScaleEngine` will be set
        '''
        if scaleDraw is None:
            scaleDraw = FancyScaleDraw()
        if scaleEngine is None:
            scaleEngine = Qwt5.QwtLinearScaleEngine()
        plot.setAxisScaleEngine(axis, scaleEngine)
        plot.setAxisScaleDraw(axis, scaleDraw)


class TaurusTimeScaleDraw(FancyScaleDraw):

    def __init__(self, *args):
        FancyScaleDraw.__init__(self, *args)

    def setDatetimeLabelFormat(self, format):
        self._datetimeLabelFormat = format

    def datetimeLabelFormat(self):
        return self._datetimeLabelFormat

    def label(self, val):
        if str(self._labelFormat) == "":
            return Qwt5.QwtText()
        # From val to a string with time
        t = datetime.fromtimestamp(val)
        try:  # If the scaleDiv was created by a DateTimeScaleEngine it has a _datetimeLabelFormat
            s = t.strftime(self._datetimeLabelFormat)
        except AttributeError:
            print("Warning: cannot get the datetime label format (Are you using a DateTimeScaleEngine?)")
            s = t.isoformat(' ')
        return Qwt5.QwtText(s)


class DeltaTimeScaleEngine(Qwt5.QwtLinearScaleEngine):

    def __init__(self, scaleDraw=None):
        Qwt5.QwtLinearScaleEngine.__init__(self)
        self.setScaleDraw(scaleDraw)

    def setScaleDraw(self, scaleDraw):
        self._scaleDraw = scaleDraw

    def scaleDraw(self):
        return self._scaleDraw

    def divideScale(self, x1, x2, maxMajSteps, maxMinSteps, stepSize):
        ''' Reimplements Qwt5.QwtLinearScaleEngine.divideScale

        :return: (Qwt5.QwtScaleDiv) a scale division whose ticks are aligned with
                 the natural delta time units '''
        interval = Qwt5.QwtDoubleInterval(x1, x2).normalized()
        if interval.width() <= 0:
            return Qwt5.QwtScaleDiv()
        d_range = interval.width()
        if d_range < 2:  # 2s
            return Qwt5.QwtLinearScaleEngine.divideScale(self, x1, x2, maxMajSteps, maxMinSteps, stepSize)
        elif d_range < 20:  # 20 s
            s = 1
        elif d_range < 120:  # =60s*2 = 2 minutes
            s = 10
        elif d_range < 1200:  # 60s*20 =20 minutes
            s = 60
        elif d_range < 7200:  # 3600s*2 = 2 hours
            s = 600
        elif d_range < 172800:  # 3600s24*2 = 2 days
            s = 3600
        else:
            s = 86400  # 1 day
        # calculate a step size that respects the base step (s) and also
        # enforces the maxMajSteps
        stepSize = s * int(numpy.ceil(float(d_range // s) / maxMajSteps))
        return Qwt5.QwtLinearScaleEngine.divideScale(self, x1, x2, maxMajSteps, maxMinSteps, stepSize)

    @staticmethod
    def getDefaultAxisLabelsAlignment(axis, rotation):
        '''return a "smart" alignment for the axis labels depending on the axis
        and the label rotation

        :param axis: (Qwt5.QwtPlot.Axis) the axis
        :param rotation: (float) The rotation (in degrees, clockwise-positive)

        :return: (Qt.Alignment) an alignment
        '''
        return _getDefaultAxisLabelsAlignment(axis, rotation)

    @staticmethod
    def enableInAxis(plot, axis, scaleDraw=None, rotation=None):
        '''convenience method that will enable this engine in the given
        axis. Note that it changes the ScaleDraw as well.

        :param plot: (Qwt5.QwtPlot) the plot to change
        :param axis: (Qwt5.QwtPlot.Axis) the id of the axis
        :param scaleDraw: (Qwt5.QwtScaleDraw) Scale draw to use. If None given,
                          the current ScaleDraw for the plot will be used if
                          possible, and a :class:`TaurusTimeScaleDraw` will be set if not
        :param rotation: (float or None) The rotation of the labels (in degrees, clockwise-positive)
        '''
        if scaleDraw is None:
            scaleDraw = plot.axisScaleDraw(axis)
            if not isinstance(scaleDraw, DeltaTimeScaleDraw):
                scaleDraw = DeltaTimeScaleDraw()
        plot.setAxisScaleDraw(axis, scaleDraw)
        plot.setAxisScaleEngine(axis, DeltaTimeScaleEngine(scaleDraw))
        if rotation is not None:
            alignment = DeltaTimeScaleEngine.getDefaultAxisLabelsAlignment(
                axis, rotation)
            plot.setAxisLabelRotation(axis, rotation)
            plot.setAxisLabelAlignment(axis, alignment)

    @staticmethod
    def disableInAxis(plot, axis, scaleDraw=None, scaleEngine=None):
        '''convenience method that will disable this engine in the given
        axis. Note that it changes the ScaleDraw as well.

        :param plot: (Qwt5.QwtPlot) the plot to change
        :param axis: (Qwt5.QwtPlot.Axis) the id of the axis
        :param scaleDraw: (Qwt5.QwtScaleDraw) Scale draw to use. If None given,
                          a :class:`FancyScaleDraw` will be set
        :param scaleEngine: (Qwt5.QwtScaleEngine) Scale draw to use. If None given,
                          a :class:`Qwt5.QwtLinearScaleEngine` will be set
        '''
        if scaleDraw is None:
            scaleDraw = FancyScaleDraw()
        if scaleEngine is None:
            scaleEngine = Qwt5.QwtLinearScaleEngine()
        plot.setAxisScaleEngine(axis, scaleEngine)
        plot.setAxisScaleDraw(axis, scaleDraw)


class DeltaTimeScaleDraw(FancyScaleDraw):

    def __init__(self, *args):
        FancyScaleDraw.__init__(self, *args)

    def label(self, val):
        if val >= 0:
            s = "+%s" % str(timedelta(seconds=val))
        else:
            s = "-%s" % str(timedelta(seconds=-val))
        return Qwt5.QwtText(s)


class FixedLabelsScaleEngine(Qwt5.QwtLinearScaleEngine):

    def __init__(self, positions):
        '''labels is a sequence of (pos,label) tuples where pos is the point
        at wich to draw the label and label is given as a python string (or QwtText)'''
        Qwt5.QwtScaleEngine.__init__(self)
        self._positions = positions
        # self.setAttribute(self.Floating,True)

    def divideScale(self, x1, x2, maxMajSteps, maxMinSteps, stepSize=0.0):
        div = Qwt5.QwtScaleDiv(x1, x2, self._positions, [], [])
        div.setTicks(Qwt5.QwtScaleDiv.MajorTick, self._positions)
        return div

    @staticmethod
    def enableInAxis(plot, axis, scaleDraw=None):
        '''convenience method that will enable this engine in the given
        axis. Note that it changes the ScaleDraw as well.

        :param plot: (Qwt5.QwtPlot) the plot to change
        :param axis: (Qwt5.QwtPlot.Axis) the id of the axis
        :param scaleDraw: (Qwt5.QwtScaleDraw) Scale draw to use. If None given,
                          the current ScaleDraw for the plot will be used if
                          possible, and a :class:`FixedLabelsScaleDraw` will be set if not
        '''
        if scaleDraw is None:
            scaleDraw = plot.axisScaleDraw(axis)
            if not isinstance(scaleDraw, FixedLabelsScaleDraw):
                scaleDraw = FixedLabelsScaleDraw()
        plot.setAxisScaleDraw(axis, scaleDraw)
        plot.setAxisScaleEngine(axis, FixedLabelsScaleEngine(scaleDraw))

    @staticmethod
    def disableInAxis(plot, axis, scaleDraw=None, scaleEngine=None):
        '''convenience method that will disable this engine in the given
        axis. Note that it changes the ScaleDraw as well.

        :param plot: (Qwt5.QwtPlot) the plot to change
        :param axis: (Qwt5.QwtPlot.Axis) the id of the axis
        :param scaleDraw: (Qwt5.QwtScaleDraw) Scale draw to use. If None given,
                          a :class:`FancyScaleDraw` will be set
        :param scaleEngine: (Qwt5.QwtScaleEngine) Scale draw to use. If None given,
                          a :class:`Qwt5.QwtLinearScaleEngine` will be set
        '''
        if scaleDraw is None:
            scaleDraw = FancyScaleDraw()
        if scaleEngine is None:
            scaleEngine = Qwt5.QwtLinearScaleEngine()
        plot.setAxisScaleEngine(axis, scaleEngine)
        plot.setAxisScaleDraw(axis, scaleDraw)


class FixedLabelsScaleDraw(FancyScaleDraw):

    def __init__(self, positions, labels):
        '''This is a custom ScaleDraw that shows labels at given positions (and nowhere else)
        positions is a sequence of points for which labels are defined.
        labels is a sequence strings (or QwtText)
        Note that the lengths of positions and labels must match'''

        if len(positions) != len(labels):
            raise ValueError('lengths of positions and labels do not match')

        FancyScaleDraw.__init__(self)
        self._positions = positions
        self._labels = labels
        # self._positionsarray = numpy.array(self._positions) #this is stored
        # just in case

    def label(self, val):
        try:
            index = self._positions.index(val)  # try to find an exact match
        except:
            index = None  # It won't show any label
            # use the index of the closest position
            #index = (numpy.abs(self._positionsarray - val)).argmin()
        if index is not None:
            return Qwt5.QwtText(self._labels[index])
        else:
            Qwt5.QwtText()



from taurus.qt.qtgui.application import TaurusApplication
from taurus.qt.qtgui.input import TaurusValueComboBox
from taurus.core.units import UR
import taurus
import pytest
from taurus.external.qt import PYSIDE2


@pytest.mark.parametrize(
    "model,names,value,expected",
    [
        # test with quantities
        ('eval:@taurus.core.evaluation.test.res.mymod.MyClass()/self.foo',
         [("A", 1234), ("B", "123"), ("C", 123 * UR.mm), ("E", -123)],
         123 * UR.mm,
         "C",
         ),
        # test with a boolean
        ('sys/tg_test/1/boolean_scalar',
         [("N", None), ("F", False), ("T", True)],
         False,
         "F",
         ),
        # test with a boolean spectrum
        ('sys/tg_test/1/boolean_spectrum',
         [
             ("A", False),
             ("B", [False, False, False]),
             ("C", [False, True, False]),
         ],
         [False, True, False],
         "C",
         ),
        # test with a string
        ('sys/tg_test/1/string_scalar',
         [("A", "foobarbaz"), ("B", "FOOBAR"), ("C", "foobar")],
         "foobar",
         "C",
         ),
        # test non-match
        ('sys/tg_test/1/string_scalar',
         [("A", "foobarbaz"), ("B", "FOOBAR"), ("C", "foobar")],
         "foo",
         "",
         ),
    ]
)
def test_TaurusValueCombobox(qtbot, model, names, value, expected):
    """Check that the TaurusValueComboBox is started with the right display
    See https://github.com/taurus-org/taurus/pull/1032
    """
    a = taurus.Attribute(model)
    a.write(value)
    w = TaurusValueComboBox()
    qtbot.addWidget(w)
    # ----------------------------------
    # workaround: avoid PySide2 segfaults when adding quantity to combobox
    # https://bugreports.qt.io/browse/PYSIDE-683
    if isinstance(value, UR.Quantity) and PYSIDE2:
        pytest.skip("avoid segfault due to PYSIDE-683 bug")
    # ----------------------------------
    w.addValueNames(names)
    qtbot.wait_until(lambda: w.count() == len(names), timeout=3200)
    try:
        with qtbot.waitSignal(w.valueChangedSignal, timeout=3200):
            w.setModel(model)
        assert w.currentText() == expected
    finally:
        del a
        # set model to None as an attempt to avoid problems in atexit()
        with qtbot.waitSignal(w.valueChangedSignal, timeout=3200):
            w.setModel(None)



from taurus.core.util.log import deprecated as __deprecated

__deprecated(dep='taurus.core.tango.search',
           alt='taurus.core.util.fandango_search', rel='4.1.2')

from taurus.core.util.fandango_search import *



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Find path of executable specified by filename"""

__all__ = ["whichfile"]

__docformat__ = "restructuredtext"


import os


def whichfile(filename, exts=None):
    '''Find path of executable specified by filename.
    It Takes into consideration
    the PATHEXT variable (found in MS Windows systems) to try also executable
    names extended with executable extensions.

    Example::

      # on a debian machine with taurus installed in the default path:
      whichfile('taurus') --> '/usr/bin/taurus'

      # or, on a winXP machine:
      whichfile('command') --> 'C:\\WINDOWS\\system32\\command.COM'

    :param filename: (str) executable name.
    :param exts: (list<str>) a list of valid executable extensions.
                 If None given, the PATHEXT environmental variable is used
                 if available.

    :return: (str) absolute path to the executable in the file system
    '''
    path = os.getenv('PATH', '').split(os.path.pathsep)
    if exts is None:
        exts = os.getenv('PATHEXT', '').split(os.path.pathsep)
    if '' not in exts:
        exts.insert(0, '')
    for p in path:
        p = os.path.join(p, filename)
        for e in exts:
            pext = p + e
            if os.access(pext, os.X_OK):
                return pext
    return None



import os


available = os.environ["AVAILABLE_QT_MOCKS"]
if available != "all" and __name__ not in available.split(","):
    raise ImportError("{} mock not enabled".format(__name__))

__version__ = 4.99



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module contains some Taurus-wide default configurations.

The idea is that the final user may edit the values here to customize certain
aspects of Taurus.
"""

#: Widget alternatives. Some widgets may be have alternative implementations
#: (e.g. the `TaurusPlot` is implemented both by the qwt5 submodule and by the
#: taurus_pyqtgraph plugin). The different implementations are registered in
#: entry point groups (`taurus.plot.alt`, `taurus.trend.alt`, ...) and they
#: are tried in alphabetical order of their registered  entry point names
#: (the first one that works is used). You can restrict the set of available
#: implementation alternatives to be tried (or even just select a given
#: alternative) by setting the corresponding *_ALT variable with a name
#: regexp pattern that must be matched by the entry point name in order to be
#: tried. For example, to force the `taurus_pyqtgraph` implementation for the
#: plots, set `PLOT_ALT = "tpg"`.
#: Leaving the variable undefined is equivalent to setting it to `".*"`
PLOT_ALT = ".*"
TREND_ALT = ".*"
TREND2D_ALT = ".*"
IMAGE_ALT = ".*"

#: Default include and exclude patterns for TaurusForm item factories
#: See `TaurusForm.setItemFactories` docs. By default, all available
#: factories are enabled (and tried alphabetically)
T_FORM_ITEM_FACTORIES = {"include": (".*",), "exclude": ()}

#: Compact mode for widgets
#: True sets the preferred mode of TaurusForms to use "compact" widgets
T_FORM_COMPACT = False

#: Strict RFC3986 URI names in models.
#: True makes Taurus only use the strict URI names
#: False enables a backwards-compatibility mode for pre-sep3 model names
STRICT_MODEL_NAMES = False

#: Lightweight imports:
#: True enables delayed imports (may break older code).
#: False (or commented out) for backwards compatibility
LIGHTWEIGHT_IMPORTS = False

#: Default scheme (if not defined, "tango" is assumed)
DEFAULT_SCHEME = "tango"

#: Filter old tango events:
#: Sometimes TangoAttribute can receive an event with an older timestamp
#: than its current one. See https://github.com/taurus-org/taurus/issues/216
#: True discards (Tango) events whose timestamp is older than the cached one.
#: False (or commented out) for backwards (pre 4.1) compatibility
FILTER_OLD_TANGO_EVENTS = True

#: Extra Taurus schemes. You can add a list of modules to be loaded for
#: providing support to new schemes
#: (e.g. EXTRA_SCHEME_MODULES = ['myownschememodule']
EXTRA_SCHEME_MODULES = []

#: Custom formatter. Taurus widgets use a default formatter based on the
#: attribute type, but sometimes a custom formatter is needed.
#: IMPORTANT: setting this option in this file will affect ALL widgets
#: of ALL applications (which is probably **not** what you want, since it
#: may have unexpected effects in some applications).
#: Consider using the API for modifying this on a per-widget or per-class
#: basis at runtime, or using the related `--default-formatter` parameter
#: from TaurusApplication, e.g.:
#:     $ taurus form MODEL --default-formatter='{:2.3f}'
#: The formatter can be a python format string or the name of a formatter
#: callable, e.g.
#: DEFAULT_FORMATTER = '{0}'
#: DEFAULT_FORMATTER = 'taurus.core.tango.util.tangoFormatter'
#: If not defined, taurus.qt.qtgui.base.defaultFormatter will be used


#: Default serialization mode **for the tango scheme**. Possible values are:
#: 'Serial', 'Concurrent', or 'TangoSerial' (default)
TANGO_SERIALIZATION_MODE = 'TangoSerial'

#: Whether TangoAttribute is subscribed to configuration events by default.
#: Setting to True (or not setting it) makes the TangoAttribute auto-subscribe
#: Setting to False avoids this subscription, which prevents issues such as
#: https://github.com/taurus-org/taurus/issues/1118
#: but it also prevents clients to be notified when configurations (e.g.,
#: units, format) change.
TANGO_AUTOSUBSCRIBE_CONF = True

#: PLY (lex/yacc) optimization: 1=Active (default) , 0=disabled.
#: Set PLY_OPTIMIZE = 0 if you are getting yacc exceptions while loading
#: synoptics
PLY_OPTIMIZE = 1

# Taurus namespace  # TODO: NAMESPACE setting seems to be unused. remove?
NAMESPACE = 'taurus'

# ----------------------------------------------------------------------------
# Qt configuration
# ----------------------------------------------------------------------------

#: Set preferred API (if one is not already loaded). Accepted values are
#: pyqt5, pyqt, pyside2, pyside. Set to an empty string to let taurus choose
#: the first that works from the accepted values.
DEFAULT_QT_API = ''

#: Auto initialize Qt logging to python logging
QT_AUTO_INIT_LOG = True

#: Remove input hook (only valid for PyQt4)
QT_AUTO_REMOVE_INPUTHOOK = True

#: Avoid application abort on unhandled python exceptions
#: (which happens since PyQt 5.5).
#: http://pyqt.sf.net/Docs/PyQt5/incompatibilities.html#unhandled-python-exceptions
#: If True (or commented out) an except hook is added to force the old
# behaviour (exception is just printed) on pyqt5
QT_AVOID_ABORT_ON_EXCEPTION = True

#: Select the theme to be used: set the theme dir  and the theme name.
#: The path can be absolute or relative to the dir of taurus.qt.qtgui.icon
#: If not set, the dir of taurus.qt.qtgui.icon will be used
QT_THEME_DIR = ''

#: The name of the icon theme (e.g. 'Tango', 'Oxygen', etc). Default='Tango'
QT_THEME_NAME = 'Tango'

#: In Linux the QT_THEME_NAME is not applied (to respect the system theme)
#: setting QT_THEME_FORCE_ON_LINUX=True overrides this.
QT_THEME_FORCE_ON_LINUX = False

#: Full Qt designer path (including filename. Default is None, meaning:
#: - linux: look for the system designer following Qt.QLibraryInfo.BinariesPath
#: - windows: look for the system designer following
#: Qt.QLibraryInfo.BinariesPath. If this fails, taurus tries to locate binary
#: manually
QT_DESIGNER_PATH = None

#: Custom organization logo. Set the absolute path to an image file to be used as your
#: organization logo. Qt registered paths can also be used.
#: If not set, it defaults to 'logos:taurus.png"
#: (note that "logos:" is a Qt a registered path for "<taurus>/qt/qtgui/icon/logos/")
ORGANIZATION_LOGO = "logos:taurus.png"

#: Implicit optparse legacy support:
#: In taurus < 4.6.5 if TaurusApplication did not receive an explicit
#: `cmd_line_parser` keyword argument, it implicitly used a
#: `optparse.OptionParser` instance. This was inconvenient because it forced
#: the user to explicitly pass `cmd_line_parser=None` when using other
#: mechanisms such as `click` or `argparse` to parse CLI options.
#: In taurus >=4.6.5 this is no longer the case by default, but the old
#: behaviour can be restored by setting IMPLICIT_OPTPARSE=True
IMPLICIT_OPTPARSE = False

# ----------------------------------------------------------------------------
# Deprecation handling:
# Note: this API is still experimental and may be subject to change
# (hence the "_" in the options)
# ----------------------------------------------------------------------------

#: set the maximum number of same-message deprecations to be logged.
#: None (or not set) indicates no limit. -1 indicates that an exception should
#: be raised instead of logging the message (useful for finding obsolete code)
_MAX_DEPRECATIONS_LOGGED = 1

# ----------------------------------------------------------------------------
# DEPRECATED SETTINGS
# ----------------------------------------------------------------------------

#: DEPRECATED. Use "taurus.form.item_factories" plugin group instead
#: A map for using custom widgets for certain devices in TaurusForms. It is a
#: dictionary with the following structure:
#: device_class_name:(classname_with_full_module_path, args, kwargs)
#: where the args and kwargs will be passed to the constructor of the class
T_FORM_CUSTOM_WIDGET_MAP = {}
_OLD_T_FORM_CUSTOM_WIDGET_MAP = {
    'SimuMotor': ('sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV', (), {}),
    'Motor': ('sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV', (), {}),
    'PseudoMotor': ('sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV', (), {}),
    'PseudoCounter': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
    'CTExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
    'ZeroDExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
    'OneDExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
    'TwoDExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
    'IORegister': ('sardana.taurus.qt.qtgui.extra_pool.PoolIORegisterTV', (), {}),
}
try:  # just for backwards compatibility. This will be removed.
    import sardana
    if sardana.release.version < '3':
        T_FORM_CUSTOM_WIDGET_MAP = _OLD_T_FORM_CUSTOM_WIDGET_MAP
except:
    pass



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ["PintValidator"]

from taurus.external.qt import Qt
from taurus.core.units import Quantity, UR
from pint import DimensionalityError


class PintValidator(Qt.QValidator):
    """A QValidator for pint Quantities"""
    _top = None
    _bottom = None
    _implicit_units = None

    @property
    def top(self):
        """
        :return: (Quantity or None) maximum accepted or None if it is not
                 enforced
        """
        return self._top

    def setTop(self, top):
        """
        Set maximum limit

        :param top: (Quantity or None) maximum acceptable value or None if it is
                    not to be enforced
        """
        self._top = Quantity(top)

    @property
    def units(self):
        """
        :return: (pint.Unit or None) base units or None if it should not
                 be enforced
        """
        return self._implicit_units

    def setUnits(self, units):
        """
        Set implicit units. They will be assumed when the text does not explicit
        the unit. They will also be used for dimensionality coherence checks.

        :param units: (pint.Unit or None). The implicit unit. If None, implicit
                      dimension is "unitless" and no dimensionality checks
                      will be performed (other than those inherent to range
                      enforcement)
        """
        self._implicit_units = units

    @property
    def bottom(self):
        """
        :return: (Quantity or None) minimum accepted or None if it is not
                 enforced
        """
        return self._bottom

    def setBottom(self, bottom):
        """
        Set minimum limit

        :param bottom: (Quantity or None) minimum acceptable value or None if it
                       is not to be enforced
        """
        self._bottom = Quantity(bottom)

    def validate(self, input, pos):
        """Reimplemented from :class:`QValidator` to validate if the input
        string is a representation of a quantity within the set bottom and top
        limits
        """
        try:
            q = Quantity(input)
        except:
            return Qt.QValidator.Intermediate, input, pos
        if self._implicit_units is not None:
            if q.unitless:
                # "cast" to implicit units
                q = Quantity(q.magnitude, self.units)
            # check coherence with implicit units
            elif self._implicit_units.dimensionality != q.dimensionality:
                return Qt.QValidator.Intermediate, input, pos
        try:
            if self.bottom is not None and q < self.bottom:
                return Qt.QValidator.Intermediate, input, pos
            if self.top is not None and q > self.top:
                return Qt.QValidator.Intermediate, input, pos
        except DimensionalityError:
            return Qt.QValidator.Intermediate, input, pos
        return Qt.QValidator.Acceptable, input, pos



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for epicsattributes..."""

import os
import sys
import numpy
import subprocess
import unittest
from taurus.core.units import Quantity
import taurus
from taurus.test import insertTest, getResourcePath
from taurus.core.taurusbasetypes import DataType, AttrQuality, DataFormat
from taurus.core.taurusbasetypes import TaurusAttrValue
import pytest




@insertTest(helper_name='write_read_attr',
            attrname='ca:test:a',
            setvalue=Quantity('1000mm'),
            expected=dict(rvalue=Quantity('1m'),
                          type=DataType.Float,
                          writable=True,
                          data_format=DataFormat._0D,
                          range=[Quantity('-10m'), Quantity('10m')],
                          alarms=[None, None],
                          warnings=[None, None],
                          ),
            expected_attrv=dict(rvalue=Quantity('1m'),
                                wvalue=Quantity('1m'),
                                quality=AttrQuality.ATTR_VALID,
                                error=None,
                                )
            )
@unittest.skipIf(('epics' in sys.modules) is False,
                 "epics module is not available")
@pytest.mark.xfail(reason='epics CV init issues need fixing in this test')
class AttributeTestCase(unittest.TestCase):
    """TestCase for the taurus.Attribute helper"""
    _process = None

    @classmethod
    def setUpClass(cls):
        """Run the epics_test softIoc"""
        db_name = getResourcePath(
            'taurus.core.epics.test.res', 'epics_test.db')
        args = ['softIoc', '-m', 'INST=test', '-d', db_name]
        dev_null = open(os.devnull, 'wb')
        cls._process = subprocess.Popen(args, stdout=dev_null, stderr=dev_null)

    @classmethod
    def tearDownClass(cls):
        """Terminate the epics_test softIoc process"""
        if cls._process:
            cls._process.terminate()
        else:
            taurus.warning('Process not started, cannot terminate it.')

    def write_read_attr(self, attrname=None, setvalue=None, expected=None,
                        expected_attrv=None, expectedshape=None):
        """check creation and correct write-and-read of an attribute"""
        if expected is None:
            expected = {}
        if expected_attrv is None:
            expected_attrv = {}

        a = taurus.Attribute(attrname)

        if setvalue is None:
            read_value = a.read(cache=False)
        else:
            read_value = a.write(setvalue,  with_read=True)

        # avoid EpicsAttribute disconnection warning
        # a.getPV().disconnect()
        msg = ('read() for "%s" did not return a TaurusAttrValue (got a %s)' %
               (attrname, read_value.__class__.__name__))
        self.assertTrue(isinstance(read_value, TaurusAttrValue), msg)

        # Test attribute
        for k, exp in expected.items():
            try:
                got = getattr(a, k)
            except AttributeError:
                msg = ('The attribute, "%s" does not provide info on %s' %
                       (attrname, k))
                self.fail(msg)
            msg = ('%s for "%s" should be %r (got %r)' %
                   (k, attrname,  exp, got))
            self.__assertValidValue(exp, got, msg)

        # Test attribute value
        for k, exp in expected_attrv.items():
            try:
                got = getattr(read_value, k)
            except AttributeError:
                msg = ('The read value for "%s" does not provide info on %s' %
                       (attrname, k))
                self.fail(msg)
            msg = ('%s for the value of %s should be %r (got %r)' %
                   (k, attrname,  exp, got))
            self.__assertValidValue(exp, got, msg)

        if expectedshape is not None:
            msg = ('rvalue.shape for %s should be %r (got %r)' %
                   (attrname, expectedshape, read_value.rvalue.shape))
            self.assertEqual(read_value.rvalue.shape, expectedshape, msg)

    def __assertValidValue(self, exp, got, msg):
        # if we are dealing with quantities, use the magnitude for comparing
        if isinstance(got, Quantity):
            got = got.to(Quantity(exp).units).magnitude
        if isinstance(exp, Quantity):
            exp = exp.magnitude
        try:
            # for those values that can be handled by numpy.allclose()
            chk = numpy.allclose(got, exp)
        except:
            if isinstance(got, numpy.ndarray):
                # uchars were not handled with allclose
                # UGLY!! but numpy.all does not work
                chk = got.tolist() == exp.tolist()
            else:
                # for the rest
                chk = bool(got == exp)

        self.assertTrue(chk, msg)


#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
curvesAppearanceChooserDlg.py:
    A Qt dialog for choosing plot appearance (symbols and lines)
    for a QwtPlot-derived widget (like Taurusplot)
"""

from __future__ import print_function

from builtins import object
import copy

from taurus.external.qt import Qt, Qwt5, compat
from taurus.core.util.containers import CaselessDict
from taurus.qt.qtgui.util.ui import UILoadable


NamedLineStyles = {None: "",
                   Qt.Qt.NoPen: "No line",
                   Qt.Qt.SolidLine: "_____",
                   Qt.Qt.DashLine: "_ _ _",
                   Qt.Qt.DotLine: ".....",
                   Qt.Qt.DashDotLine: "_._._",
                   Qt.Qt.DashDotDotLine: ".._..",
                   }
ReverseNamedLineStyles = {}
for k, v in NamedLineStyles.items():
    ReverseNamedLineStyles[v] = k

NamedCurveStyles = {None: "",
                    Qwt5.QwtPlotCurve.NoCurve: "No curve",
                    Qwt5.QwtPlotCurve.Lines: "Lines",
                    Qwt5.QwtPlotCurve.Sticks: "Sticks",
                    Qwt5.QwtPlotCurve.Steps: "Steps",
                    Qwt5.QwtPlotCurve.Dots: "Dots"
                    }
ReverseNamedCurveStyles = {}
for k, v in NamedCurveStyles.items():
    ReverseNamedCurveStyles[v] = k

NamedSymbolStyles = {
    None: "",
    Qwt5.QwtSymbol.NoSymbol: "No symbol",
    Qwt5.QwtSymbol.Ellipse: "Circle",
    Qwt5.QwtSymbol.Rect: "Square",
    Qwt5.QwtSymbol.Diamond: "Diamond",
    Qwt5.QwtSymbol.Triangle: "Triangle",
    Qwt5.QwtSymbol.DTriangle: "Down Triangle",
    Qwt5.QwtSymbol.UTriangle: "Up triangle",
    Qwt5.QwtSymbol.LTriangle: "Left Triangle",
    Qwt5.QwtSymbol.RTriangle: "Right Triangle",
    Qwt5.QwtSymbol.Cross: "Cross",
    Qwt5.QwtSymbol.XCross: "XCross",
    Qwt5.QwtSymbol.HLine: "Horizontal line",
    Qwt5.QwtSymbol.VLine: "Vertical line",
    Qwt5.QwtSymbol.Star1: "Star1",
    Qwt5.QwtSymbol.Star2: "Star2",
    Qwt5.QwtSymbol.Hexagon: "Hexagon"
}

ReverseNamedSymbolStyles = {}
for k, v in NamedSymbolStyles.items():
    ReverseNamedSymbolStyles[v] = k

NamedColors = ["Black", "Red", "Blue", "Magenta",
               "Green", "Cyan", "Yellow", "Gray", "White"]


@UILoadable
class CurvesAppearanceChooser(Qt.QWidget):
    """
    A widget for choosing plot appearance for one or more curves.
    The current curves properties are passed using the setCurves() method using
    a dictionary with the following structure::

        curvePropDict={name1:prop1, name2:prop2,...}

    where propX is an instance of :class:`CurveAppearanceProperties`
    When applying, a signal is emitted and the chosen properties are made
    available in a similar dictionary. """

    NAME_ROLE = Qt.Qt.UserRole

    controlChanged = Qt.pyqtSignal()
    curveAppearanceChanged = Qt.pyqtSignal(compat.PY_OBJECT, list)
    CurveTitleEdited = Qt.pyqtSignal('QString', 'QString')

    def __init__(self, parent=None, curvePropDict={}, showButtons=False, autoApply=False, designMode=False):
        # try:
        super(CurvesAppearanceChooser, self).__init__(parent)
        self.loadUi()
        self.autoApply = autoApply
        self.sStyleCB.insertItems(0, sorted(NamedSymbolStyles.values()))
        self.lStyleCB.insertItems(0, list(NamedLineStyles.values()))
        self.cStyleCB.insertItems(0, list(NamedCurveStyles.values()))
        self.sColorCB.addItem("")
        self.lColorCB.addItem("")
        if not showButtons:
            self.applyBT.hide()
            self.resetBT.hide()
        for color in NamedColors:
            icon = self._colorIcon(color)
            self.sColorCB.addItem(icon, "", Qt.QColor(color))
            self.lColorCB.addItem(icon, "", Qt.QColor(color))
        self.__itemsDict = CaselessDict()
        self.setCurves(curvePropDict)
        # set the icon for the background button (stupid designer limitations
        # forces to do it programatically)
        self.bckgndBT.setIcon(Qt.QIcon(":color-fill.svg"))

        # connections.
        # Note: The assignToY1BT and assignToY2BT buttons are not connected to anything
        # Their signals are handled by the Config dialog because we haven't got
        # access to the curve objects here
        self.curvesLW.itemSelectionChanged.connect(self.onSelectedCurveChanged)
        self.curvesLW.itemChanged.connect(self.onItemChanged)
        self.applyBT.clicked.connect(self.onApply)
        self.resetBT.clicked.connect(self.onReset)
        self.sStyleCB.currentIndexChanged.connect(self._onSymbolStyleChanged)

        self.sStyleCB.currentIndexChanged.connect(self.onControlChanged)
        self.lStyleCB.currentIndexChanged.connect(self.onControlChanged)
        self.sColorCB.currentIndexChanged.connect(self.onControlChanged)
        self.lColorCB.currentIndexChanged.connect(self.onControlChanged)
        self.cStyleCB.currentIndexChanged.connect(self.onControlChanged)
        self.sSizeSB.valueChanged.connect(self.onControlChanged)
        self.lWidthSB.valueChanged.connect(self.onControlChanged)
        self.sFillCB.stateChanged.connect(self.onControlChanged)
        self.cFillCB.stateChanged.connect(self.onControlChanged)
        # except Exception, e:
        # print "CURVE APPEARANCE EXCEPTION:",str(e)

    def setCurves(self, curvePropDict):
        '''Populates the list of curves from the properties dictionary. It uses
        the curve title for display, and stores the curve name as the item data
        (with role=CurvesAppearanceChooser.NAME_ROLE)

        :param curvePropDict:   (dict) a dictionary whith keys=curvenames and
                                values= :class:`CurveAppearanceProperties` object
        '''
        self.curvePropDict = curvePropDict
        self._curvePropDictOrig = copy.deepcopy(curvePropDict)
        self.curvesLW.clear()
        self.__itemsDict = CaselessDict()
        for name, prop in self.curvePropDict.items():
            # create and insert the item
            item = Qt.QListWidgetItem(str(prop.title), self.curvesLW)
            self.__itemsDict[name] = item
            item.setData(self.NAME_ROLE, str(name))
            item.setToolTip("<b>Curve Name:</b> %s" % name)
            item.setFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsSelectable |
                          Qt.Qt.ItemIsUserCheckable | Qt.Qt.ItemIsDragEnabled | Qt.Qt.ItemIsEditable)
        self.curvesLW.setCurrentRow(0)

    def onItemChanged(self, item):
        '''slot used when an item data has changed'''
        name = item.data(self.NAME_ROLE)
        previousTitle = self.curvePropDict[name].title
        currentTitle = item.text()
        if previousTitle != currentTitle:
            self.curvePropDict[name].title = currentTitle
            self.CurveTitleEdited.emit(name, currentTitle)

    def updateTitles(self, newTitlesDict=None):
        '''
        Updates the titles of the curves that are displayed in the curves list.

        :param newTitlesDict: (dict<str,str>) dictionary with key=curve_name and
                              value=title
        '''
        if newTitlesDict is None:
            return
        for name, title in newTitlesDict.items():
            self.curvePropDict[name].title = title
            self.__itemsDict[name].setText(title)

    def getSelectedCurveNames(self):
        '''Returns the curve names for the curves selected at the curves list.

        *Note*: The names may differ from the displayed text, which
        corresponds to the curve titles (this method is what you likely need if
        you want to get keys to use in curves or curveProp dicts).

        :return: (string_list) the names of the selected curves
        '''
        return [item.data(self.NAME_ROLE) for item in self.curvesLW.selectedItems()]

    def showProperties(self, prop=None):
        '''Updates the dialog to show the given properties.

        :param prop: (CurveAppearanceProperties) the properties object
                     containing what should be shown. If a given property is set
                     to None, the corresponding widget will show a "neutral"
                     display
        '''
        if prop is None:
            prop = self._shownProp
        # set the Style comboboxes
        self.sStyleCB.setCurrentIndex(
            self.sStyleCB.findText(NamedSymbolStyles[prop.sStyle]))
        self.lStyleCB.setCurrentIndex(
            self.lStyleCB.findText(NamedLineStyles[prop.lStyle]))
        self.cStyleCB.setCurrentIndex(
            self.cStyleCB.findText(NamedCurveStyles[prop.cStyle]))
        # set sSize and lWidth spinboxes. if prop.sSize is None, it puts -1
        # (which is the special value for these switchhboxes)
        self.sSizeSB.setValue(max(prop.sSize, -1))
        self.lWidthSB.setValue(max(prop.lWidth, -1))
        # Set the Color combo boxes. The item at index 0 is the empty one in
        # the comboboxes Manage unknown colors by including them
        if prop.sColor is None:
            index = 0
        else:
            index = self.sColorCB.findData(Qt.QColor(prop.sColor))
        if index == -1:  # if the color is not one of the supported colors, add it to the combobox
            index = self.sColorCB.count()  # set the index to what will be the added one
            self.sColorCB.addItem(self._colorIcon(Qt.QColor(prop.sColor)),
                                  "",
                                  Qt.QColor(prop.sColor)
                                  )
        self.sColorCB.setCurrentIndex(index)
        if prop.lColor is None:
            index = 0
        else:
            index = self.lColorCB.findData(Qt.QColor(prop.lColor))
        if index == -1:  # if the color is not one of the supported colors, add it to the combobox
            index = self.lColorCB.count()  # set the index to what will be the added one
            self.lColorCB.addItem(self._colorIcon(Qt.QColor(prop.lColor)),
                                  "",
                                  Qt.QColor(prop.lColor)
                                  )
        self.lColorCB.setCurrentIndex(index)
        # set the Fill Checkbox. The prop.sFill value can be in 3 states: True,
        # False and None
        if prop.sFill is None:
            checkState = Qt.Qt.PartiallyChecked
        elif prop.sFill:
            checkState = Qt.Qt.Checked
        else:
            checkState = Qt.Qt.Unchecked
        # set the Area Fill Checkbox. The prop.cFill value can be in 3 states:
        # True, False and None
        if prop.cFill is None:
            checkState = Qt.Qt.PartiallyChecked
        elif prop.cFill:
            checkState = Qt.Qt.Checked
        else:
            checkState = Qt.Qt.Unchecked
        self.cFillCB.setCheckState(checkState)

    def onControlChanged(self, *args):
        '''slot to be called whenever a control widget is changed. It emmits a
        'controlChanged signal and applies the change if in autoapply mode.
        It ignores any arguments passed'''

        self.controlChanged.emit()
        if self.autoApply:
            self.onApply()

    def onSelectedCurveChanged(self):
        """Updates the shown properties when the curve selection changes"""
        plist = [self.curvePropDict[name]
                 for name in self.getSelectedCurveNames()]
        if len(plist) == 0:
            plist = [CurveAppearanceProperties()]
        self._shownProp = CurveAppearanceProperties.merge(plist)
        self.showProperties(self._shownProp)

    def _onSymbolStyleChanged(self, text):
        '''Slot called when the Symbol style is changed, to ensure that symbols
        are visible if you choose them

        :param text: (str) the new symbol style label
        '''
        text = str(text)
        if self.sSizeSB.value() < 2 and not text in ["", "No symbol"]:
            # a symbol size of 0 is invisible and 1 means you should use
            # cStyle=dots
            self.sSizeSB.setValue(3)

    def getShownProperties(self):
        """Returns a copy of the currently shown properties and updates
        self._shownProp

        :return: (CurveAppearanceProperties)
        """
        prop = CurveAppearanceProperties()
        # get the values from the Style comboboxes. Note that the empty string
        # ("") translates into None
        prop.sStyle = ReverseNamedSymbolStyles[
            str(self.sStyleCB.currentText())]
        prop.lStyle = ReverseNamedLineStyles[str(self.lStyleCB.currentText())]
        prop.cStyle = ReverseNamedCurveStyles[str(self.cStyleCB.currentText())]
        # get sSize and lWidth from the spinboxes
        prop.sSize = self.sSizeSB.value()
        prop.lWidth = self.lWidthSB.value()
        if prop.sSize < 0:
            prop.sSize = None
        if prop.lWidth < 0:
            prop.lWidth = None
        # Get the Color combo boxes. The item at index 0 is the empty one in
        # the comboboxes
        index = self.sColorCB.currentIndex()
        if index == 0:
            prop.sColor = None
        else:
            prop.sColor = Qt.QColor(self.sColorCB.itemData(index))
        index = self.lColorCB.currentIndex()
        if index == 0:
            prop.lColor = None
        else:
            prop.lColor = Qt.QColor(self.lColorCB.itemData(index))
        # get the sFill from the Checkbox.
        checkState = self.sFillCB.checkState()
        if checkState == Qt.Qt.PartiallyChecked:
            prop.sFill = None
        else:
            prop.sFill = bool(checkState)
        # get the cFill from the Checkbox.
        checkState = self.cFillCB.checkState()
        if checkState == Qt.Qt.PartiallyChecked:
            prop.cFill = None
        else:
            prop.cFill = bool(checkState)
        # store the props
        self._shownProp = copy.deepcopy(prop)
        return copy.deepcopy(prop)

    def onApply(self):
        """Apply does 2 things:

            - It updates `self.curvePropDict` using the current values
              choosen in the dialog
            - It emits a curveAppearanceChanged signal that indicates the names
              of the curves that changed and the new properties. (The names and
              the properties are returned by the function as well)

        :return: (tuple<CurveAppearanceProperties,list>) a tuple containing the
                 curve properties and a list of the selected curve names (as a
                 list<str>)
        """
        names = self.getSelectedCurveNames()
        prop = self.getShownProperties()
        # Update self.curvePropDict for selected properties
        for n in names:
            self.curvePropDict[n] = CurveAppearanceProperties.merge([self.curvePropDict[n], prop],
                                                                    conflict=CurveAppearanceProperties.inConflict_update_a)
        # emit a (PyQt) signal telling what properties (first argument) need to
        # be applied to which curves (second argument)
        self.curveAppearanceChanged.emit(prop, names)
        # return both values
        return prop, names

    def onReset(self):
        '''slot to be called when the reset action is triggered. It reverts to
        the original situation'''
        self.setCurves(self._curvePropDictOrig)
        self.curvesLW.clearSelection()

    def _colorIcon(self, color, w=10, h=10):
        # to do: create a border
        pixmap = Qt.QPixmap(w, h)
        pixmap.fill(Qt.QColor(color))
        return Qt.QIcon(pixmap)


class CurveAppearanceProperties(object):
    '''An object describing the appearance of a TaurusCurve'''

    def __init__(self, sStyle=None, sSize=None, sColor=None, sFill=None,
                 lStyle=None, lWidth=None, lColor=None, cStyle=None,
                 yAxis=None, cFill=None, title=None, visible=None):
        """
        Creator of :class:`CurveAppearanceProperties`
        Possible keyword arguments are:
            - sStyle= symbolstyle
            - sSize= int
            - sColor= color
            - sFill= bool
            - lStyle= linestyle
            - lWidth= int
            - lColor= color
            - cStyle= curvestyle
            - cFill= bool
            - yAxis= axis
            - visible = bool
            - title= title

        Where:
            - color is a color that QColor() understands (i.e. a
              Qt.Qt.GlobalColor, a color name, or a Qt.Qcolor)
            - symbolstyle is one of Qwt5.QwtSymbol.Style
            - linestyle is one of Qt.Qt.PenStyle
            - curvestyle is one of Qwt5.QwtPlotCurve.CurveStyle
            - axis is one of Qwt5.QwtPlot.Axis
            - title is something that Qwt5.QwtText() accepts in its constructor
              (i.e. a QwtText or a string type)
        """
        self.sStyle = sStyle
        self.sSize = sSize
        self.sColor = sColor
        self.sFill = sFill
        self.lStyle = lStyle
        self.lWidth = lWidth
        self.lColor = lColor
        self.cStyle = cStyle
        self.cFill = cFill
        self.yAxis = yAxis
        self.title = title
        self.visible = visible
        self.propertyList = ["sStyle", "sSize", "sColor", "sFill", "lStyle", "lWidth",
                             "lColor", "cStyle", "cFill", "yAxis", "title", "visible"]

    def _print(self):
        """Just for debug"""
        print("-" * 77)
        for k in self.propertyList:
            print(k + "= ", self.__getattribute__(k))
        print("-" * 77)

    @staticmethod
    def inConflict_update_a(a, b):
        """This  function can be passed to CurvesAppearance.merge()
        if one wants to update prop1 with prop2 except for those
        attributes of prop2 that are set to None"""
        if b is None:
            return a
        else:
            return b

    @staticmethod
    def inConflict_none(a, b):
        """In case of conflict, returns None"""
        return None

    def conflictsWith(self, other, strict=True):
        """returns a list of attribute names that are in conflict between this self and other"""
        result = []
        for aname in self.propertyList:
            vself = getattr(self, aname)
            vother = getattr(other, aname)
            if (vself != vother) and (strict or not(vself is None or vother is None)):
                result.append(aname)
        return result

    @classmethod
    def merge(self, plist, attributes=None, conflict=None):
        """returns a CurveAppearanceProperties object formed by merging a list
        of other CurveAppearanceProperties objects

        **Note:** This is a class method, so it can be called without previously
        instantiating an object

        :param plist: (sequence<CurveAppearanceProperties>) objects to be merged
        :param attributes: (sequence<str>) the name of the attributes to
                           consider for the merge. If None, all the attributes
                           will be merged
        :param conflict: (callable) a function that takes 2 objects (having a
                         different attribute)and returns a value that solves the
                         conflict. If None is given, any conflicting attribute
                         will be set to None.

        :return: (CurveAppearanceProperties) merged properties
        """

        n = len(plist)
        if n < 1:
            raise ValueError("plist must contain at least 1 member")
        plist = copy.deepcopy(plist)
        if n == 1:
            return plist[0]
        if attributes is None:
            attributes = ["sStyle", "sSize", "sColor", "sFill", "lStyle",
                          "lWidth", "lColor", "cStyle", "cFill", "yAxis", "title"]
        if conflict is None:
            conflict = CurveAppearanceProperties.inConflict_none
        p = CurveAppearanceProperties()
        for a in attributes:
            alist = [p.__getattribute__(a) for p in plist]
            p.__setattr__(a, alist[0])
            for ai in alist[1:]:
                if alist[0] != ai:
                    # print "MERGING:",alist[0],ai,conflict(alist[0],ai)
                    p.__setattr__(a, conflict(alist[0], ai))
                    break
        return p

    def applyToCurve(self, curve):
        """applies the current properties to a given curve
        If a property is set to None, it is not applied to the curve"""
        raise DeprecationWarning(
            "CurveAppearanceProperties.applyToCurve() is deprecated. Use TaurusCurve.setAppearanceProperties() instead")
        curve.setAppearanceProperties(self)

#        s=curve.symbol()
#        if self.sStyle is not None: s.setStyle(symbol[self.sStyle])
#        if self.sSize is not None: s.setSize(self.sSize)
#        if self.sColor is not None: s.brush().setColor(Qt.QColor(self.sColor))
#        if self.sFill is not None:
#            if self.sFill: s.brush().setStyle(Qt.Qt.SolidPattern)
#            else: s.brush().setStyle(Qt.Qt.NoBrush)
#        p=curve.pen()
#        if self.lStyle is not None: p.setStyle(lineStyles[self.lStyle])
#        if self.lWidth is not None: p.setWidth(self.lWidth)
#        if self.lColor is not None: p.setColor(Qt.QColor(self.lColor))
#        curveStyle=curve.style()
#        if self.cStyle is not None: curveStyle.setStyle(self.cStyle)
#        if self.cFill is not None:
#            if self.cFill:
#                color = p.color()
#                color.setAlphaF(0.5)
#                b = self.brush()
#                b.setColor(color)
#                b.setStyle(Qt.Qt.SolidPattern)
#            else:
#                c.brush().setStyle(Qt.Qt.NoBrush)
#        if self.yAxis is not None: curve.setYAxis(self.yAxis)
#        if self.title is not None: curve.setTitle(Qwt5.QwtText(self.title))



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from __future__ import print_function
print("*" * 77)
print("""
If you are seeing this, it is because you tried to access qwtplot.py directly.
All funcionality has been moved to taurusplot.py, taurustrend.py and scales.py

The recommended way of accessing the classes that were in qwtplot.py is by
importing them from taurus.qt.qtgui.plot

If you were trying to launch a stand-alone Taurusplot or TaurusTrend from a script,
you should update such script.
""")
print("*" * 77)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a demo for the :class:`~taurus.qt.qtgui.display.TaurusLCD`
widget """

__all__ = ["demo", "main"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt


def demo():
    import sys
    import taurus.qt.qtgui.application
    import taurus.qt.qtgui.display

    Application = taurus.qt.qtgui.application.TaurusApplication
    TaurusLCD = taurus.qt.qtgui.display.TaurusLCD

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application()

    M = 2

    class TaurusLCDTestPanel(Qt.QWidget):

        def __init__(self, parent=None):
            Qt.QWidget.__init__(self, parent)
            panel_l = Qt.QVBoxLayout()
            self.setLayout(panel_l)
            panel_l.setContentsMargins(M, M, M, M)
            panel_l.setSpacing(M)

            w = TaurusLCD()
            display_panel = Qt.QGroupBox("Taurus LCD Display")
            display_l = Qt.QHBoxLayout()
            display_l.setContentsMargins(M, M, M, M)
            display_l.setSpacing(M)
            display_panel.setLayout(display_l)
            display_l.addWidget(w, 1)

            control_panel = Qt.QGroupBox("Control Panel")
            control_l = Qt.QFormLayout()
            control_l.setContentsMargins(M, M, M, M)
            control_l.setSpacing(M)
            control_panel.setLayout(control_l)
            model_widget = Qt.QLineEdit()
            model_index_widget = Qt.QLineEdit()
            fg_widget = Qt.QComboBox()
            bg_widget = Qt.QComboBox()
            control_l.addRow("model:", model_widget)
            control_l.addRow("model index:", model_index_widget)
            control_l.addRow("foreground role:", fg_widget)
            control_l.addRow("background role:", bg_widget)

            panel_l.addWidget(display_panel)
            panel_l.addWidget(control_panel)

            fg_widget.addItems(["value", "w_value", "none"])
            bg_widget.addItems(["quality", "state", "none"])

            model_widget.textChanged.connect(w.setModel)
            model_index_widget.textChanged.connect(w.setModelIndex)
            fg_widget.currentIndexChanged['QString'].connect(w.setFgRole)
            bg_widget.currentIndexChanged['QString'].connect(w.setBgRole)

            model_widget.setText("sys/tg_test/1/double_scalar")
            fg_widget.setCurrentIndex(0)
            bg_widget.setCurrentIndex(0)

            self.w_label = w
            self.w_model = model_widget
            self.w_model_index = model_index_widget
            self.w_fg = fg_widget
            self.w_bg = bg_widget

    panel = Qt.QWidget()
    layout = Qt.QGridLayout()
    panel.setLayout(layout)
    layout.setContentsMargins(M, M, M, M)
    layout.setSpacing(M)
    p1 = TaurusLCDTestPanel()
    p1.w_model.setText("sys/tg_test/1/double_scalar")
    p2 = TaurusLCDTestPanel()
    p2.w_model.setText("sys/tg_test/1/double_scalar#min_value")
    p2.w_bg.setCurrentIndex(2)
    layout.addWidget(p1, 0, 0)
    layout.addWidget(p2, 0, 1)
    #layout.addItem(Qt.QSpacerItem(10,10), 1, 0, 1, 2)
    layout.setRowStretch(0, 1)
    panel.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return panel


def main():
    return demo()

if __name__ == "__main__":
    main()



import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)
panel = Qt.QWidget()
layout = Qt.QHBoxLayout()
panel.setLayout(layout)

from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.input import TaurusValueLineEdit, TaurusValueSpinBox, TaurusWheelEdit

w1 = TaurusLabel()
w2 = TaurusLabel()
w3 = TaurusValueLineEdit()  # or TaurusValueSpinBox or TaurusWheelEdit
w4 = TaurusLabel()
layout.addWidget(w1)
layout.addWidget(w2)
layout.addWidget(w3)
layout.addWidget(w4)
w1.model, w1.bgRole = 'sys/tg_test/1/double_scalar#label', ''
w2.model = 'sys/tg_test/1/double_scalar#rvalue.magnitude'
w3.model = 'sys/tg_test/1/double_scalar#rvalue.magnitude'
w4.model, w4.bgRole = 'sys/tg_test/1/double_scalar#unit', ''

panel.show()
sys.exit(app.exec_())



'''
This code exposes a limitation in the useParentModel architecture of Taurus and
explains how to avoid/work-around it.

Note that the conditions for hitting this limitation are quite special and that
in most situations you will not hit it. But unfortunately it may happen with
designer-generated files, and then the recheckTaurusParent() workaround is useful:
call recheckTaurusParent for all designer created widgets that use TaurusParentModel.
You can do it right after calling the setupUi method.
'''
from __future__ import print_function
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.display import TaurusLabel
import sys
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)

# The problem arises in some situations when the Taurus parenting is not the same
# as the Qt parenting. For demonstration we use 3 widgets:
# p <-- m <--c (the arrows indicate the *Qt* parent)
# note that "m" not being a Taurus widget, implies that the *Taurus* parent of c is p
# also note that we are not giving the parent in the constructor, but we rely on
# doing it later on when adding to layout

p = TaurusWidget()  # Taurus parent
m = Qt.QWidget()  # midle widget (non Taurus)
c = TaurusLabel()  # Taurus child

# here we call setUseParentModel before the parent is known!

c.setUseParentModel(True)

# we prepare the layouts

m.setLayout(Qt.QVBoxLayout())
p.setLayout(Qt.QVBoxLayout())

# Now, by adding the widgets to the layout we are actually reparenting them.
# The order in which we reparent determines success/failure:
# if we do m-->p and then c-->m, it works, but if we do it in the opposite
# order, we trigger the error.
# (i.e., if we had called "p.layout().addWidget(m)" it would work work)
m.layout().addWidget(c)
p.layout().addWidget(m)

# the problem arises because the Taurus ancestry of c is only checked when:
# a) c.setUseParentModel() is called (and it effectively changes something) or
# b) when c is Qt-reparented
##
# In this example, c is reparented before m, so when it checks for a Taurus
# ancestor, none is found. Then m is reparented and c now would find its right
# Taurus ancestor (p)... but c never gets notified!

# Manually calling c.recheckTaurusParent() after the parenting has changed is
# another way to work around this issue (you can test by uncommenting
# the following line):

# c.recheckTaurusParent()

p.setModel('sys/tg_test/1/state')
print('p model:', p.getModelName())
print('c model:', c.getModelName())


p.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Tests for taurus.core.tango.tangofactory"""


import time
import taurus

from taurus.external.unittest import TestCase
from taurus.core.tango.test.tgtestds import TangoSchemeTestLauncher


class TestFactoryGarbageCollection(TangoSchemeTestLauncher, TestCase):

    # in order to not interfere with the following tests this device should
    # not be used in another tests
    DEV_NAME = 'TangoSchemeTest/unittest/temp-tfgc-1'

    def setUp(self):
        self.factory = taurus.Factory()

    # TODO: Uncomment this test when tango factory recycles authority
    # def test_authority(self):
    #     old_len = len(self.factory.tango_db)
    # 
    #     def create():
    #         taurus.Authority()
    # 
    #     create()
    #     msg = "factory is polluted with authority "
    #     self.assertEqual(len(self.factory.tango_db), old_len, msg)

    def test_device(self):
        old_len = len(self.factory.tango_devs)

        def create():
            taurus.Device(self.DEV_NAME)

        create()
        msg = "factory is polluted with device"
        self.assertEqual(len(self.factory.tango_devs), old_len, msg)

    def test_attribute(self):
        old_len = len(self.factory.tango_attrs)

        def create():
            taurus.Attribute(self.DEV_NAME + "/state")

        create()
        msg = "factory is polluted with attribute"
        self.assertEqual(len(self.factory.tango_attrs), old_len, msg)

    def tearDown(self):
        self.factory = None


def test_cleanup_after_polling():
    """
    Ensure that polling a Tango attribute does not keep device alive
    See Bug #999
    (Also check case insensitivity)
    """
    polling_period = .1  # seconds
    a = taurus.Attribute('sys/TG_test/1/FLOAT_scalar')
    f = a.factory()
    a.activatePolling(polling_period * 1000, force=True)
    assert len(list(f.tango_attrs.keys())) == 1
    assert len(list(f.tango_devs.keys())) == 1
    a = None
    time.sleep(polling_period)
    assert len(list(f.tango_attrs.keys())) == 0
    assert len(list(f.tango_devs.keys())) == 0


def test_cleanup_state_after_polling():
    """
    Ensure that polling the state Tango attribute does not keep device alive
    See Bug #999
    (Also check case insensitivity)
    """
    polling_period = .1  # seconds
    a = taurus.Attribute('sys/TG_TEST/1/STate')
    f = a.factory()
    a.activatePolling(polling_period * 1000, force=True)
    assert len(list(f.tango_attrs.keys())) == 1
    assert len(list(f.tango_devs.keys())) == 1
    a = None
    time.sleep(polling_period)
    assert len(list(f.tango_attrs.keys())) == 0
    assert len(list(f.tango_devs.keys())) == 0

